# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmExperimentEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_by': 'OrmUser',
        'models': 'list[OrmModel]',
        'plot_volume': 'OrmVolume',
        'project': 'OrmProject',
        'sweep': 'OrmSweep',
        'tags': 'list[OrmTag]',
        'workloads': 'list[OrmWorkload]'
    }

    attribute_map = {
        'created_by': 'created_by',
        'models': 'models',
        'plot_volume': 'plot_volume',
        'project': 'project',
        'sweep': 'sweep',
        'tags': 'tags',
        'workloads': 'workloads'
    }

    def __init__(self, created_by=None, models=None, plot_volume=None, project=None, sweep=None, tags=None, workloads=None, local_vars_configuration=None):  # noqa: E501
        """OrmExperimentEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_by = None
        self._models = None
        self._plot_volume = None
        self._project = None
        self._sweep = None
        self._tags = None
        self._workloads = None
        self.discriminator = None

        if created_by is not None:
            self.created_by = created_by
        if models is not None:
            self.models = models
        if plot_volume is not None:
            self.plot_volume = plot_volume
        if project is not None:
            self.project = project
        if sweep is not None:
            self.sweep = sweep
        if tags is not None:
            self.tags = tags
        if workloads is not None:
            self.workloads = workloads

    @property
    def created_by(self):
        """Gets the created_by of this OrmExperimentEdges.  # noqa: E501


        :return: The created_by of this OrmExperimentEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this OrmExperimentEdges.


        :param created_by: The created_by of this OrmExperimentEdges.  # noqa: E501
        :type created_by: OrmUser
        """

        self._created_by = created_by

    @property
    def models(self):
        """Gets the models of this OrmExperimentEdges.  # noqa: E501


        :return: The models of this OrmExperimentEdges.  # noqa: E501
        :rtype: list[OrmModel]
        """
        return self._models

    @models.setter
    def models(self, models):
        """Sets the models of this OrmExperimentEdges.


        :param models: The models of this OrmExperimentEdges.  # noqa: E501
        :type models: list[OrmModel]
        """

        self._models = models

    @property
    def plot_volume(self):
        """Gets the plot_volume of this OrmExperimentEdges.  # noqa: E501


        :return: The plot_volume of this OrmExperimentEdges.  # noqa: E501
        :rtype: OrmVolume
        """
        return self._plot_volume

    @plot_volume.setter
    def plot_volume(self, plot_volume):
        """Sets the plot_volume of this OrmExperimentEdges.


        :param plot_volume: The plot_volume of this OrmExperimentEdges.  # noqa: E501
        :type plot_volume: OrmVolume
        """

        self._plot_volume = plot_volume

    @property
    def project(self):
        """Gets the project of this OrmExperimentEdges.  # noqa: E501


        :return: The project of this OrmExperimentEdges.  # noqa: E501
        :rtype: OrmProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this OrmExperimentEdges.


        :param project: The project of this OrmExperimentEdges.  # noqa: E501
        :type project: OrmProject
        """

        self._project = project

    @property
    def sweep(self):
        """Gets the sweep of this OrmExperimentEdges.  # noqa: E501


        :return: The sweep of this OrmExperimentEdges.  # noqa: E501
        :rtype: OrmSweep
        """
        return self._sweep

    @sweep.setter
    def sweep(self, sweep):
        """Sets the sweep of this OrmExperimentEdges.


        :param sweep: The sweep of this OrmExperimentEdges.  # noqa: E501
        :type sweep: OrmSweep
        """

        self._sweep = sweep

    @property
    def tags(self):
        """Gets the tags of this OrmExperimentEdges.  # noqa: E501


        :return: The tags of this OrmExperimentEdges.  # noqa: E501
        :rtype: list[OrmTag]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """Sets the tags of this OrmExperimentEdges.


        :param tags: The tags of this OrmExperimentEdges.  # noqa: E501
        :type tags: list[OrmTag]
        """

        self._tags = tags

    @property
    def workloads(self):
        """Gets the workloads of this OrmExperimentEdges.  # noqa: E501


        :return: The workloads of this OrmExperimentEdges.  # noqa: E501
        :rtype: list[OrmWorkload]
        """
        return self._workloads

    @workloads.setter
    def workloads(self, workloads):
        """Sets the workloads of this OrmExperimentEdges.


        :param workloads: The workloads of this OrmExperimentEdges.  # noqa: E501
        :type workloads: list[OrmWorkload]
        """

        self._workloads = workloads

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmExperimentEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmExperimentEdges):
            return True

        return self.to_dict() != other.to_dict()



import atexit
import os
import re
import signal
import sys
import threading
import time
from typing import Any, Dict, List, Optional, Union

from urllib3.exceptions import MaxRetryError

from openapi_client.models.experiment_metric_entry import ExperimentMetricEntry
from openapi_client.models.experiment_metrics_update_api_payload import (
    ExperimentMetricsUpdateAPIPayload,
)
from openapi_client.models.experiment_progress_update_api_payload import (
    ExperimentProgressUpdateAPIPayload,
)
from openapi_client.models.local_experiment_finish_api_payload import (
    LocalExperimentFinishAPIPayload,
)
from openapi_client.models.response_experiment_info import ResponseExperimentInfo
from vessl.cli._util import Endpoint
from vessl.internal.collector import (
    Collector,
    IOCollector,
    SystemMetricCollector,
    UserMetricCollector,
)
from vessl.util import logger
from vessl.util.api import VesslApi
from vessl.util.audio import Audio
from vessl.util.constant import (
    VESSL_AUDIO_PATH,
    VESSL_IMAGE_PATH,
    VESSL_MEDIA_PATH,
    VESSL_PLOTS_FILETYPE_AUDIO,
    VESSL_PLOTS_FILETYPE_IMAGE,
)
from vessl.util.exception import (
    InvalidExperimentError,
    InvalidTypeError,
    VesslApiException,
)
from vessl.util.image import Image

MODE_TEST = "TEST"
MODE_NOT_STARTED = "NOT_STARTED"
MODE_LOCAL = "LOCAL"
MODE_MANAGED = "MANAGED"

SEND_INTERVAL_IN_SEC = 10
PROGRESS_UPDATE_INTERVAL_IN_SEC = 1

METRIC_KEY_REGEX = re.compile("^[a-zA-Z0-9/_-]+$")


class Sender(object):
    def __init__(self, api: VesslApi, experiment_id: int, collectors: List[Collector]):
        self._api = api
        self._experiment_id: int = experiment_id
        self._thread = threading.Thread(target=self._thread_body, daemon=True)
        self._exit = threading.Event()
        self._collectors = collectors

    def stop(self):
        for c in self._collectors:
            c.stop()

        self._exit.set()
        self._thread.join()

    def start(self):
        for c in self._collectors:
            c.start()
        self._thread.start()

    def _thread_body(self):
        while not self._exit.is_set():
            self._send()
            self._exit.wait(timeout=SEND_INTERVAL_IN_SEC)
        self._send()

    def _send(self):
        pairs = [(c, c.collect()) for c in self._collectors]
        for c, m in pairs:
            logger.debug(f"{c} / {m}", str(c), len(m))
        payload = [m for _, metrics in pairs for m in metrics]
        logger.debug(f"Sending {len(payload)} payloads")

        try:
            res = self._api.experiment_metrics_update_api(
                self._experiment_id,
                experiment_metrics_update_api_payload=ExperimentMetricsUpdateAPIPayload(
                    metrics=payload
                ),
            )
            if res.rejected:
                logger.warning(f"{res.rejected} payloads(s) were rejected.")

            for c, m in pairs:
                c.truncate(len(m))

        except (MaxRetryError, VesslApiException) as e:
            logger.exception("Failed to send metrics to server", exc_info=e)
        except Exception as e:
            logger.exception("Unexpected error", exc_info=e)


class ProgressUpdater(object):
    def __init__(self, api: VesslApi, experiment_id: int):
        self._api = api
        self._experiment_id = experiment_id
        self._thread = threading.Thread(target=self._thread_body, daemon=True)
        self._exit = threading.Event()

        self._progress = None

    def start(self):
        self._thread.start()

    def stop(self):
        self._exit.set()
        self._thread.join()

    def update(self, value: float):
        self._progress = value

    def _thread_body(self):
        while not self._exit.is_set():
            self._send()
            self._exit.wait(timeout=PROGRESS_UPDATE_INTERVAL_IN_SEC)
        self._send()

    def _send(self):
        if self._progress is None:
            return

        logger.debug(f"Sending experiment progress: {self._progress}")
        try:
            self._api.experiment_progress_update_api(
                self._experiment_id,
                experiment_progress_update_api_payload=ExperimentProgressUpdateAPIPayload(
                    progress_percent=self._progress,
                ),
            )
            self._progress = None  # Flush after sending

        except (MaxRetryError, VesslApiException) as e:
            logger.exception("Failed to send metrics to server", exc_info=e)

        except Exception as e:
            logger.exception("Unexpected error", exc_info=e)


class VesslRun(object):
    class ExitHook(object):
        def __init__(self, orig_exit):
            self.orig_exit = orig_exit
            self.exit_code = 0

        def exit(self, code=0):
            self.exit_code = code
            self.orig_exit(code)

    __slots__ = [
        "api",
        "_mode",
        "_collectors",
        "_sender",
        "_progress_updater",
        "_experiment",
        "_logger",
        "_user_metric_collector",
        "_exit_hook",
    ]

    def __init__(self) -> None:
        self.api = VesslApi()
        self._experiment = self._get_experiment_from_environment()
        self._mode = MODE_NOT_STARTED if self._experiment is None else MODE_MANAGED

        self._user_metric_collector = UserMetricCollector()
        self._exit_hook = self.ExitHook(sys.exit)

    def _get_experiment_from_environment(self) -> Optional[ResponseExperimentInfo]:
        """Detect experiment from environment variables

        In a Vessl-managed experiment, these variables will be defined.
        """
        experiment_id = os.environ.get("VESSL_EXPERIMENT_ID", None)
        access_token = os.environ.get("VESSL_ACCESS_TOKEN", None)

        if experiment_id is None or access_token is None:
            return None

        self.api.configure_access_token(access_token)
        try:
            return self.api.experiment_read_by_idapi(experiment_id=experiment_id)
        except VesslApiException:
            return None

    def _get_experiment_from_args(
        self, experiment_name_or_number, message: str = None
    ) -> ResponseExperimentInfo:
        """Get or create a local experiment

        If experiment is specified, use it. Otherwise, create a new experiment.
        """
        # Create a new experiment
        if experiment_name_or_number is None:
            from vessl.experiment import create_local_experiment

            experiment = create_local_experiment(message=message)
            logger.debug(f"Created experiment {experiment.id}")
            return experiment

        # Continue with previous experiment
        from vessl.experiment import read_experiment

        experiment = read_experiment(experiment_name_or_number)
        if not experiment.is_local or experiment.local_execution_spec is None:
            raise InvalidExperimentError(
                f"{experiment.name}: cannot use Vessl-managed experiment."
            )
        if experiment.status != "running":
            raise InvalidExperimentError(
                f"{experiment.name}: experiment must be running."
            )

        return experiment

    def _signal_handler(self, signo, frames):
        sys.exit(130)  # job was terminated by the owner

    def _start(self):
        """Start sender and register hooks"""
        self._sender.start()

        sys.exit = self._exit_hook.exit
        atexit.register(self._stop)
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)

    def _stop(self):
        """Stop sender and restore hooks"""
        if self._mode != MODE_LOCAL:
            return

        self._sender.stop()

        sys.exit = self._exit_hook.orig_exit
        self.api.local_experiment_finish_api(
            self._experiment.id,
            local_experiment_finish_api_payload=LocalExperimentFinishAPIPayload(
                exit_code=self._exit_hook.exit_code
            ),
        )

    def _send_without_collector(self, payloads):
        """Send metrics without using the collector

        In a Vessl-managed experiment, metrics are sent immediately instead of being queued.
        """
        assert self._mode == MODE_MANAGED
        res = self.api.experiment_metrics_update_api(
            self._experiment.id,
            experiment_metrics_update_api_payload=ExperimentMetricsUpdateAPIPayload(
                metrics=payloads
            ),
        )
        if res.rejected:
            logger.warning(f"{res.rejected} payloads(s) were rejected.")

    def init(self, experiment_name_or_number=None, message: str = None, is_test=False):
        """Main function to setup Vessl in a local setting

        If this is a Vessl-managed experiment or vessl.init has already been called,
        this will do nothing.

        Args:
            experiment_name_or_number (str | int): experiment name or number
            message (str): experiment message
            is_test (bool): internal use only
        """
        if self._mode != MODE_NOT_STARTED:
            return

        print(f"Initializing a new experiment...")
        self.api.configure()
        self._experiment = self._get_experiment_from_args(
            experiment_name_or_number, message
        )
        self._mode = MODE_LOCAL

        print(
            f"Connected to {self._experiment.name}.\n"
            f"For more info: {Endpoint.experiment.format(self._experiment.organization.name, self._experiment.project.name, self._experiment.number)}"
        )

        if is_test:
            return

        gpu_count = self._experiment.local_execution_spec.gpu_count or 0
        self._user_metric_collector = UserMetricCollector()
        collectors = [
            IOCollector(),
            SystemMetricCollector(gpu_count),
            self._user_metric_collector,
        ]
        self._sender = Sender(self.api, self._experiment.id, collectors)
        self._start()

    def upload(self, path: str):
        """Upload output files

        Args:
            path (str): path to upload
        """
        if self._mode == MODE_NOT_STARTED:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        from vessl.experiment import upload_experiment_output_files

        upload_experiment_output_files(self._experiment.name, path)

    def finish(self):
        """Teardown Vessl settings

        Use this function to stop tracking your experiment mid-script. If not called,
        tracking is stopped automatically upon exit.

        Args:
            path (str): path to upload
        """
        if self._mode == MODE_NOT_STARTED:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        if self._mode == MODE_MANAGED:
            return

        self._stop()
        experiment_name = self._experiment.name
        self._mode = MODE_NOT_STARTED
        self._experiment = None
        print(f"Experiment {experiment_name} completed.")

    def log(self, payload: Dict[str, Any], step: Optional[int] = None):
        """Log metrics to Vessl

        Args:
            payload (Dict[str, Any]): to log a scalar, value should be a number. To
                log an image, pass a single image or a list of images (type `vessl.util.image.Image`).
            step (int): step.
        """
        if self._mode == MODE_NOT_STARTED:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        scalar_dict = {}
        media_dict = {}

        for k, v in payload.items():
            if isinstance(v, list):
                if all(isinstance(i, Image) for i in v) or all(
                    isinstance(i, Audio) for i in v
                ):
                    media_dict[k] = v
            elif isinstance(v, Image) or isinstance(v, Audio):
                media_dict[k] = [v]
            else:
                scalar_dict[k] = v

        # Update step if step is specified. If a scalar is defined but step wasn't,
        # step will be autoincremented.
        if scalar_dict or step is not None:
            self._user_metric_collector.handle_step(step)

        if media_dict:
            self._update_media(media_dict)

        if scalar_dict:
            self._update_metrics(scalar_dict)

    # This should mirror app/influx/metric_schema.go > `isValidMetricKey`
    def _is_metric_key_valid(self, key: str):
        if not METRIC_KEY_REGEX.match(key):
            return False
        if key.startswith("/") or key.endswith("/"):
            return False
        for i in range(1, len(key)):
            if key[i] == "/" and key[i - 1] == "/":
                return False
        return True

    def _update_metrics(self, payload: Dict[str, Any]):
        invalid_keys = [k for k in payload.keys() if not self._is_metric_key_valid(k)]
        if invalid_keys:
            logger.warning(
                f"Invalid metric keys: {' '.join(invalid_keys)}. This payload will be rejected."
            )

        payloads = [self._user_metric_collector.build_metric_payload(payload)]
        if self._mode == MODE_MANAGED:
            self._send_without_collector(payloads)
            return self._user_metric_collector.step
        return self._user_metric_collector.log_metrics(payloads)

    def _update_media(self, payload: Dict[str, Union[List[Image], List[Audio]]]):
        media_type, set_media_type = None, False
        path_to_caption = {}
        for media in payload.values():
            for medium in media:
                path_to_caption[medium.path] = medium.caption
                if not set_media_type:
                    if not isinstance(medium, Image) and not isinstance(medium, Audio):
                        raise InvalidTypeError(
                            f"Invalid payload type error: {type(medium)}"
                        )
                    else:
                        media_type = type(medium).__name__
                        set_media_type = True

        # Source image path: vessl-media/images/
        # Source audio path: vessl-media/audio/
        if media_type == Image.__name__:
            source_path = os.path.join(VESSL_MEDIA_PATH, VESSL_IMAGE_PATH, "")
        elif media_type == Audio.__name__:
            source_path = os.path.join(VESSL_MEDIA_PATH, VESSL_AUDIO_PATH, "")

        assert self._experiment
        dest_volume_id = self._experiment.experiment_plot_volume
        dest_path = "/" + VESSL_MEDIA_PATH

        from vessl.volume import copy_volume_file

        files = copy_volume_file(
            source_volume_id=None,
            source_path=source_path,
            dest_volume_id=dest_volume_id,
            dest_path=dest_path,
            recursive=True,
        )

        for media in payload.values():
            for medium in media:
                medium.flush()

        if files:
            plot_files = [
                {
                    "step": None,
                    "path": file.path,
                    "caption": path_to_caption[file.path],
                    "timestamp": time.time(),
                }
                for file in files
                if file.path in path_to_caption
            ]
        else:
            plot_files = []

        payloads: List[ExperimentMetricEntry] = []
        for f in plot_files:
            payload = {}
            if media_type == Image.__name__:
                payload = {VESSL_PLOTS_FILETYPE_IMAGE: f}
            elif media_type == Audio.__name__:
                payload = {VESSL_PLOTS_FILETYPE_AUDIO: f}
            payloads.append(self._user_metric_collector.build_media_payload(payload))

        if self._mode == MODE_MANAGED:
            self._send_without_collector(payloads)
        else:
            self._user_metric_collector.log_media(payloads)

    def progress(self, value: float):
        """Update experiment progress

        Args:
            value (float): progress value as a decimal between 0 and 1
        """
        if self._experiment is None:
            logger.warning("Invalid. Use `vessl.init()` first.")
            return

        if not 0 < value <= 1:
            logger.warning(f"Invalid progress value {value}. (0 < value <= 1)")
            return

        if not hasattr(self, "_progress_updater"):
            # Do not initialize in init() since ProgressUpdater might not be used at all
            self._progress_updater = ProgressUpdater(self.api, self._experiment.id)
            self._progress_updater.start()

        self._progress_updater.update(value)
        logger.debug(f"Experiment progress: {value}")



from typing import List

from openapi_client import SavviHubDatasetCreateAPIPayload
from openapi_client.models import (
    GSDatasetCreateAPIPayload,
    ResponseDatasetInfo,
    ResponseDatasetInfoDetail,
    ResponseDatasetVersionInfo,
    ResponseFileMetadata,
    S3DatasetCreateAPIPayload,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.util.constant import DATASET_PATH_SCHEME_GS, DATASET_PATH_SCHEME_S3
from vessl.util.exception import InvalidDatasetError
from vessl.volume import copy_volume_file, delete_volume_file, list_volume_files


def read_dataset(dataset_name: str, **kwargs) -> ResponseDatasetInfoDetail:
    """Read dataset

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.dataset_read_api(
        dataset_name=dataset_name, organization_name=_get_organization_name(**kwargs)
    )


def read_dataset_version(
    dataset_id: int, dataset_version_hash: str
) -> ResponseDatasetVersionInfo:
    return vessl_api.dataset_version_read_api(
        dataset_id=dataset_id,
        dataset_version_hash=dataset_version_hash,
    )


def list_datasets(**kwargs) -> List[ResponseDatasetInfo]:
    """List datasets

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.dataset_list_api(
        organization_name=_get_organization_name(**kwargs),
    ).results


def _create_dataset_local(
    dataset_name: str,
    is_version_enabled: bool = False,
    description: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    return vessl_api.savvi_hub_dataset_create_api(
        organization_name=_get_organization_name(**kwargs),
        savvi_hub_dataset_create_api_payload=SavviHubDatasetCreateAPIPayload(
            name=dataset_name,
            description=description,
            is_version_enabled=is_version_enabled,
        ),
    )


def _create_dataset_s3(
    dataset_name: str,
    is_version_enabled: bool = False,
    description: str = None,
    external_path: str = None,
    aws_role_arn: str = None,
    version_path: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    if is_version_enabled and (
        version_path is None or not version_path.startswith(DATASET_PATH_SCHEME_S3)
    ):
        raise InvalidDatasetError(f"Invalid version path: {version_path}")

    return vessl_api.s3_dataset_create_api(
        organization_name=_get_organization_name(**kwargs),
        s3_dataset_create_api_payload=S3DatasetCreateAPIPayload(
            name=dataset_name,
            description=description,
            is_version_enabled=is_version_enabled,
            s3_path=external_path,
            version_s3_path=version_path,
            is_public=False,
            aws_role_arn=aws_role_arn,
        ),
    )


def _create_dataset_gs(
    dataset_name: str,
    is_version_enabled: bool = False,
    description: str = None,
    external_path: str = None,
    version_path: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    if is_version_enabled:
        raise InvalidDatasetError("Versioning is not supported for GoogleStorage")

    return vessl_api.g_s_dataset_create_api(
        organization_name=_get_organization_name(**kwargs),
        gs_dataset_create_api_payload=GSDatasetCreateAPIPayload(
            name=dataset_name,
            description=description,
            is_version_enabled=is_version_enabled,
            gs_path=external_path,
            version_gs_path=version_path,
            is_public=False,
        ),
    )


def create_dataset(
    dataset_name: str,
    description: str = None,
    is_version_enabled: bool = False,
    external_path: str = None,
    aws_role_arn: str = None,
    version_path: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    """Create dataset

    Keyword args:
        organization_name (str): override default organization
    """
    if external_path is None:
        return _create_dataset_local(
            dataset_name, is_version_enabled, description, **kwargs
        )

    if external_path.startswith(DATASET_PATH_SCHEME_S3):
        return _create_dataset_s3(
            dataset_name,
            is_version_enabled,
            description,
            external_path,
            aws_role_arn,
            version_path,
            **kwargs,
        )

    if external_path.startswith(DATASET_PATH_SCHEME_GS):
        return _create_dataset_gs(
            dataset_name,
            is_version_enabled,
            description,
            external_path,
            version_path,
            **kwargs,
        )

    raise InvalidDatasetError("Invalid path scheme. Must be either s3:// or gs://.")


def list_dataset_volume_files(
    dataset_name: str,
    need_download_url: bool = False,
    path: str = "",
    recursive: bool = False,
    **kwargs,
) -> List[ResponseFileMetadata]:
    """List dataset volume files

    Keyword args:
        organization_name (str): override default organization
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return list_volume_files(dataset.volume_id, need_download_url, path, recursive)


def upload_dataset_volume_file(
    dataset_name: str,
    source_path: str,
    dest_path: str,
    **kwargs,
) -> None:
    """Upload file to dataset

    Keyword args:
        organization_name (str): override default organization
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return copy_volume_file(
        source_volume_id=None,
        source_path=source_path,
        dest_volume_id=dataset.volume_id,
        dest_path=dest_path,
    )


def download_dataset_volume_file(
    dataset_name: str,
    source_path: str,
    dest_path: str,
    **kwargs,
) -> None:
    """Download file from dataset

    Keyword args:
        organization_name (str): override default organization
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return copy_volume_file(
        source_volume_id=dataset.volume_id,
        source_path=source_path,
        dest_volume_id=None,
        dest_path=dest_path,
    )


def copy_dataset_volume_file(
    dataset_name: str,
    source_path: str,
    dest_path: str,
    recursive: bool = False,
    **kwargs,
) -> None:
    """Copy files within a same dataset

    Keyword args:
        organization_name (str): override default organization
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return copy_volume_file(
        source_volume_id=dataset.volume_id,
        source_path=source_path,
        dest_volume_id=dataset.volume_id,
        dest_path=dest_path,
        recursive=recursive,
    )


def delete_dataset_volume_file(
    dataset_name: str,
    path: str,
    recursive: bool = False,
    **kwargs,
) -> List[ResponseFileMetadata]:
    """Delete dataset volume file

    Keyword args:
        organization_name (str): override default organization
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return delete_volume_file(dataset.volume_id, path, recursive)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseExperimentListResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_by': 'ResponseSimpleUser',
        'created_dt': 'datetime',
        'env_vars': 'OrmEnvVars',
        'estimated_end_dt': 'datetime',
        'histories': 'list[ResponseWorkloadHistoryInfo]',
        'id': 'int',
        'is_distributed': 'bool',
        'is_local': 'bool',
        'kernel_cluster': 'ResponseKernelCluster',
        'kernel_resource_spec': 'ResponseKernelResourceSpec',
        'last_alive': 'datetime',
        'message': 'str',
        'metrics_summary': 'OrmExperimentMetricsSummary',
        'name': 'str',
        'number': 'int',
        'organization': 'ResponseOrganization',
        'progress_percent': 'float',
        'project': 'ResponseProject',
        'show_plots': 'bool',
        'status': 'str',
        'status_last_updated': 'datetime',
        'status_reason': 'str',
        'sweep': 'ResponseSimpleSweep',
        'tags': 'list[ResponseTagResponse]',
        'updated_dt': 'datetime',
        'workload_status': 'str'
    }

    attribute_map = {
        'created_by': 'created_by',
        'created_dt': 'created_dt',
        'env_vars': 'env_vars',
        'estimated_end_dt': 'estimated_end_dt',
        'histories': 'histories',
        'id': 'id',
        'is_distributed': 'is_distributed',
        'is_local': 'is_local',
        'kernel_cluster': 'kernel_cluster',
        'kernel_resource_spec': 'kernel_resource_spec',
        'last_alive': 'last_alive',
        'message': 'message',
        'metrics_summary': 'metrics_summary',
        'name': 'name',
        'number': 'number',
        'organization': 'organization',
        'progress_percent': 'progress_percent',
        'project': 'project',
        'show_plots': 'show_plots',
        'status': 'status',
        'status_last_updated': 'status_last_updated',
        'status_reason': 'status_reason',
        'sweep': 'sweep',
        'tags': 'tags',
        'updated_dt': 'updated_dt',
        'workload_status': 'workload_status'
    }

    def __init__(self, created_by=None, created_dt=None, env_vars=None, estimated_end_dt=None, histories=None, id=None, is_distributed=None, is_local=None, kernel_cluster=None, kernel_resource_spec=None, last_alive=None, message=None, metrics_summary=None, name=None, number=None, organization=None, progress_percent=None, project=None, show_plots=None, status=None, status_last_updated=None, status_reason=None, sweep=None, tags=None, updated_dt=None, workload_status=None, local_vars_configuration=None):  # noqa: E501
        """ResponseExperimentListResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_by = None
        self._created_dt = None
        self._env_vars = None
        self._estimated_end_dt = None
        self._histories = None
        self._id = None
        self._is_distributed = None
        self._is_local = None
        self._kernel_cluster = None
        self._kernel_resource_spec = None
        self._last_alive = None
        self._message = None
        self._metrics_summary = None
        self._name = None
        self._number = None
        self._organization = None
        self._progress_percent = None
        self._project = None
        self._show_plots = None
        self._status = None
        self._status_last_updated = None
        self._status_reason = None
        self._sweep = None
        self._tags = None
        self._updated_dt = None
        self._workload_status = None
        self.discriminator = None

        self.created_by = created_by
        self.created_dt = created_dt
        if env_vars is not None:
            self.env_vars = env_vars
        self.estimated_end_dt = estimated_end_dt
        self.histories = histories
        self.id = id
        self.is_distributed = is_distributed
        self.is_local = is_local
        if kernel_cluster is not None:
            self.kernel_cluster = kernel_cluster
        if kernel_resource_spec is not None:
            self.kernel_resource_spec = kernel_resource_spec
        if last_alive is not None:
            self.last_alive = last_alive
        self.message = message
        if metrics_summary is not None:
            self.metrics_summary = metrics_summary
        self.name = name
        self.number = number
        self.organization = organization
        self.progress_percent = progress_percent
        self.project = project
        self.show_plots = show_plots
        self.status = status
        self.status_last_updated = status_last_updated
        self.status_reason = status_reason
        if sweep is not None:
            self.sweep = sweep
        self.tags = tags
        self.updated_dt = updated_dt
        if workload_status is not None:
            self.workload_status = workload_status

    @property
    def created_by(self):
        """Gets the created_by of this ResponseExperimentListResponse.  # noqa: E501


        :return: The created_by of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: ResponseSimpleUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ResponseExperimentListResponse.


        :param created_by: The created_by of this ResponseExperimentListResponse.  # noqa: E501
        :type created_by: ResponseSimpleUser
        """
        if self.local_vars_configuration.client_side_validation and created_by is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by`, must not be `None`")  # noqa: E501

        self._created_by = created_by

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseExperimentListResponse.  # noqa: E501


        :return: The created_dt of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseExperimentListResponse.


        :param created_dt: The created_dt of this ResponseExperimentListResponse.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def env_vars(self):
        """Gets the env_vars of this ResponseExperimentListResponse.  # noqa: E501


        :return: The env_vars of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: OrmEnvVars
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this ResponseExperimentListResponse.


        :param env_vars: The env_vars of this ResponseExperimentListResponse.  # noqa: E501
        :type env_vars: OrmEnvVars
        """

        self._env_vars = env_vars

    @property
    def estimated_end_dt(self):
        """Gets the estimated_end_dt of this ResponseExperimentListResponse.  # noqa: E501


        :return: The estimated_end_dt of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._estimated_end_dt

    @estimated_end_dt.setter
    def estimated_end_dt(self, estimated_end_dt):
        """Sets the estimated_end_dt of this ResponseExperimentListResponse.


        :param estimated_end_dt: The estimated_end_dt of this ResponseExperimentListResponse.  # noqa: E501
        :type estimated_end_dt: datetime
        """

        self._estimated_end_dt = estimated_end_dt

    @property
    def histories(self):
        """Gets the histories of this ResponseExperimentListResponse.  # noqa: E501


        :return: The histories of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: list[ResponseWorkloadHistoryInfo]
        """
        return self._histories

    @histories.setter
    def histories(self, histories):
        """Sets the histories of this ResponseExperimentListResponse.


        :param histories: The histories of this ResponseExperimentListResponse.  # noqa: E501
        :type histories: list[ResponseWorkloadHistoryInfo]
        """
        if self.local_vars_configuration.client_side_validation and histories is None:  # noqa: E501
            raise ValueError("Invalid value for `histories`, must not be `None`")  # noqa: E501

        self._histories = histories

    @property
    def id(self):
        """Gets the id of this ResponseExperimentListResponse.  # noqa: E501


        :return: The id of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseExperimentListResponse.


        :param id: The id of this ResponseExperimentListResponse.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def is_distributed(self):
        """Gets the is_distributed of this ResponseExperimentListResponse.  # noqa: E501


        :return: The is_distributed of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: bool
        """
        return self._is_distributed

    @is_distributed.setter
    def is_distributed(self, is_distributed):
        """Sets the is_distributed of this ResponseExperimentListResponse.


        :param is_distributed: The is_distributed of this ResponseExperimentListResponse.  # noqa: E501
        :type is_distributed: bool
        """
        if self.local_vars_configuration.client_side_validation and is_distributed is None:  # noqa: E501
            raise ValueError("Invalid value for `is_distributed`, must not be `None`")  # noqa: E501

        self._is_distributed = is_distributed

    @property
    def is_local(self):
        """Gets the is_local of this ResponseExperimentListResponse.  # noqa: E501


        :return: The is_local of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: bool
        """
        return self._is_local

    @is_local.setter
    def is_local(self, is_local):
        """Sets the is_local of this ResponseExperimentListResponse.


        :param is_local: The is_local of this ResponseExperimentListResponse.  # noqa: E501
        :type is_local: bool
        """
        if self.local_vars_configuration.client_side_validation and is_local is None:  # noqa: E501
            raise ValueError("Invalid value for `is_local`, must not be `None`")  # noqa: E501

        self._is_local = is_local

    @property
    def kernel_cluster(self):
        """Gets the kernel_cluster of this ResponseExperimentListResponse.  # noqa: E501


        :return: The kernel_cluster of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: ResponseKernelCluster
        """
        return self._kernel_cluster

    @kernel_cluster.setter
    def kernel_cluster(self, kernel_cluster):
        """Sets the kernel_cluster of this ResponseExperimentListResponse.


        :param kernel_cluster: The kernel_cluster of this ResponseExperimentListResponse.  # noqa: E501
        :type kernel_cluster: ResponseKernelCluster
        """

        self._kernel_cluster = kernel_cluster

    @property
    def kernel_resource_spec(self):
        """Gets the kernel_resource_spec of this ResponseExperimentListResponse.  # noqa: E501


        :return: The kernel_resource_spec of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: ResponseKernelResourceSpec
        """
        return self._kernel_resource_spec

    @kernel_resource_spec.setter
    def kernel_resource_spec(self, kernel_resource_spec):
        """Sets the kernel_resource_spec of this ResponseExperimentListResponse.


        :param kernel_resource_spec: The kernel_resource_spec of this ResponseExperimentListResponse.  # noqa: E501
        :type kernel_resource_spec: ResponseKernelResourceSpec
        """

        self._kernel_resource_spec = kernel_resource_spec

    @property
    def last_alive(self):
        """Gets the last_alive of this ResponseExperimentListResponse.  # noqa: E501


        :return: The last_alive of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._last_alive

    @last_alive.setter
    def last_alive(self, last_alive):
        """Sets the last_alive of this ResponseExperimentListResponse.


        :param last_alive: The last_alive of this ResponseExperimentListResponse.  # noqa: E501
        :type last_alive: datetime
        """

        self._last_alive = last_alive

    @property
    def message(self):
        """Gets the message of this ResponseExperimentListResponse.  # noqa: E501


        :return: The message of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this ResponseExperimentListResponse.


        :param message: The message of this ResponseExperimentListResponse.  # noqa: E501
        :type message: str
        """
        if self.local_vars_configuration.client_side_validation and message is None:  # noqa: E501
            raise ValueError("Invalid value for `message`, must not be `None`")  # noqa: E501

        self._message = message

    @property
    def metrics_summary(self):
        """Gets the metrics_summary of this ResponseExperimentListResponse.  # noqa: E501


        :return: The metrics_summary of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: OrmExperimentMetricsSummary
        """
        return self._metrics_summary

    @metrics_summary.setter
    def metrics_summary(self, metrics_summary):
        """Sets the metrics_summary of this ResponseExperimentListResponse.


        :param metrics_summary: The metrics_summary of this ResponseExperimentListResponse.  # noqa: E501
        :type metrics_summary: OrmExperimentMetricsSummary
        """

        self._metrics_summary = metrics_summary

    @property
    def name(self):
        """Gets the name of this ResponseExperimentListResponse.  # noqa: E501


        :return: The name of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseExperimentListResponse.


        :param name: The name of this ResponseExperimentListResponse.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def number(self):
        """Gets the number of this ResponseExperimentListResponse.  # noqa: E501


        :return: The number of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: int
        """
        return self._number

    @number.setter
    def number(self, number):
        """Sets the number of this ResponseExperimentListResponse.


        :param number: The number of this ResponseExperimentListResponse.  # noqa: E501
        :type number: int
        """
        if self.local_vars_configuration.client_side_validation and number is None:  # noqa: E501
            raise ValueError("Invalid value for `number`, must not be `None`")  # noqa: E501

        self._number = number

    @property
    def organization(self):
        """Gets the organization of this ResponseExperimentListResponse.  # noqa: E501


        :return: The organization of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: ResponseOrganization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this ResponseExperimentListResponse.


        :param organization: The organization of this ResponseExperimentListResponse.  # noqa: E501
        :type organization: ResponseOrganization
        """
        if self.local_vars_configuration.client_side_validation and organization is None:  # noqa: E501
            raise ValueError("Invalid value for `organization`, must not be `None`")  # noqa: E501

        self._organization = organization

    @property
    def progress_percent(self):
        """Gets the progress_percent of this ResponseExperimentListResponse.  # noqa: E501


        :return: The progress_percent of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: float
        """
        return self._progress_percent

    @progress_percent.setter
    def progress_percent(self, progress_percent):
        """Sets the progress_percent of this ResponseExperimentListResponse.


        :param progress_percent: The progress_percent of this ResponseExperimentListResponse.  # noqa: E501
        :type progress_percent: float
        """

        self._progress_percent = progress_percent

    @property
    def project(self):
        """Gets the project of this ResponseExperimentListResponse.  # noqa: E501


        :return: The project of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: ResponseProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this ResponseExperimentListResponse.


        :param project: The project of this ResponseExperimentListResponse.  # noqa: E501
        :type project: ResponseProject
        """
        if self.local_vars_configuration.client_side_validation and project is None:  # noqa: E501
            raise ValueError("Invalid value for `project`, must not be `None`")  # noqa: E501

        self._project = project

    @property
    def show_plots(self):
        """Gets the show_plots of this ResponseExperimentListResponse.  # noqa: E501


        :return: The show_plots of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: bool
        """
        return self._show_plots

    @show_plots.setter
    def show_plots(self, show_plots):
        """Sets the show_plots of this ResponseExperimentListResponse.


        :param show_plots: The show_plots of this ResponseExperimentListResponse.  # noqa: E501
        :type show_plots: bool
        """
        if self.local_vars_configuration.client_side_validation and show_plots is None:  # noqa: E501
            raise ValueError("Invalid value for `show_plots`, must not be `None`")  # noqa: E501

        self._show_plots = show_plots

    @property
    def status(self):
        """Gets the status of this ResponseExperimentListResponse.  # noqa: E501


        :return: The status of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ResponseExperimentListResponse.


        :param status: The status of this ResponseExperimentListResponse.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def status_last_updated(self):
        """Gets the status_last_updated of this ResponseExperimentListResponse.  # noqa: E501


        :return: The status_last_updated of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._status_last_updated

    @status_last_updated.setter
    def status_last_updated(self, status_last_updated):
        """Sets the status_last_updated of this ResponseExperimentListResponse.


        :param status_last_updated: The status_last_updated of this ResponseExperimentListResponse.  # noqa: E501
        :type status_last_updated: datetime
        """
        if self.local_vars_configuration.client_side_validation and status_last_updated is None:  # noqa: E501
            raise ValueError("Invalid value for `status_last_updated`, must not be `None`")  # noqa: E501

        self._status_last_updated = status_last_updated

    @property
    def status_reason(self):
        """Gets the status_reason of this ResponseExperimentListResponse.  # noqa: E501


        :return: The status_reason of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: str
        """
        return self._status_reason

    @status_reason.setter
    def status_reason(self, status_reason):
        """Sets the status_reason of this ResponseExperimentListResponse.


        :param status_reason: The status_reason of this ResponseExperimentListResponse.  # noqa: E501
        :type status_reason: str
        """
        if self.local_vars_configuration.client_side_validation and status_reason is None:  # noqa: E501
            raise ValueError("Invalid value for `status_reason`, must not be `None`")  # noqa: E501

        self._status_reason = status_reason

    @property
    def sweep(self):
        """Gets the sweep of this ResponseExperimentListResponse.  # noqa: E501


        :return: The sweep of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: ResponseSimpleSweep
        """
        return self._sweep

    @sweep.setter
    def sweep(self, sweep):
        """Sets the sweep of this ResponseExperimentListResponse.


        :param sweep: The sweep of this ResponseExperimentListResponse.  # noqa: E501
        :type sweep: ResponseSimpleSweep
        """

        self._sweep = sweep

    @property
    def tags(self):
        """Gets the tags of this ResponseExperimentListResponse.  # noqa: E501


        :return: The tags of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: list[ResponseTagResponse]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """Sets the tags of this ResponseExperimentListResponse.


        :param tags: The tags of this ResponseExperimentListResponse.  # noqa: E501
        :type tags: list[ResponseTagResponse]
        """
        if self.local_vars_configuration.client_side_validation and tags is None:  # noqa: E501
            raise ValueError("Invalid value for `tags`, must not be `None`")  # noqa: E501

        self._tags = tags

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseExperimentListResponse.  # noqa: E501


        :return: The updated_dt of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseExperimentListResponse.


        :param updated_dt: The updated_dt of this ResponseExperimentListResponse.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    @property
    def workload_status(self):
        """Gets the workload_status of this ResponseExperimentListResponse.  # noqa: E501


        :return: The workload_status of this ResponseExperimentListResponse.  # noqa: E501
        :rtype: str
        """
        return self._workload_status

    @workload_status.setter
    def workload_status(self, workload_status):
        """Sets the workload_status of this ResponseExperimentListResponse.


        :param workload_status: The workload_status of this ResponseExperimentListResponse.  # noqa: E501
        :type workload_status: str
        """

        self._workload_status = workload_status

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseExperimentListResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseExperimentListResponse):
            return True

        return self.to_dict() != other.to_dict()



import os
import platform
import sys
from typing import Dict, List

import psutil
import pynvml

from openapi_client import DistributedExperimentCreateAPIPayload
from openapi_client.models import (
    ExperimentCreateAPIPayload,
    InfluxdbWorkloadLog,
    LocalExperimentCreateAPIPayload,
    OrmEnvVar,
    ResponseExperimentInfo,
    ResponseExperimentListResponse,
    ResponseFileMetadata,
)
from openapi_client.models.orm_execution_environment import OrmExecutionEnvironment
from vessl import __version__, vessl_api
from vessl.kernel_cluster import read_cluster
from vessl.kernel_resource_spec import (
    _configure_custom_kernel_resource_spec,
    read_kernel_resource_spec,
)
from vessl.organization import _get_organization_name
from vessl.project import _get_project_name
from vessl.util import logger
from vessl.util.common import safe_cast
from vessl.util.constant import MOUNT_PATH_OUTPUT, MOUNT_TYPE_OUTPUT
from vessl.util.exception import InvalidExperimentError, InvalidParamsError
from vessl.volume import (
    _configure_volume_mount_requests,
    copy_volume_file,
    list_volume_files,
)


def read_experiment(experiment_name_or_number: str, **kwargs) -> ResponseExperimentInfo:
    """Read experiment

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    return vessl_api.experiment_read_api(
        experiment=experiment_name_or_number,
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
    )


def read_experiment_by_id(experiment_id: int, **kwargs) -> ResponseExperimentInfo:
    result = vessl_api.experiment_read_by_idapi(experiment_id=experiment_id)
    assert isinstance(result, ResponseExperimentInfo)
    return result


def list_experiments(**kwargs) -> List[ResponseExperimentListResponse]:
    """List experiments

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    return vessl_api.experiment_list_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
    ).results


def create_experiment(
    cluster_name: str,
    start_command: str,
    kernel_resource_spec_name: str = None,
    processor_type: str = None,
    cpu_limit: float = None,
    memory_limit: str = None,
    gpu_type: str = None,
    gpu_limit: int = None,
    kernel_image_url: str = None,
    *,
    message: str = None,
    termination_protection: bool = False,
    env_vars: List[str] = None,
    dataset_mounts: List[str] = None,
    git_ref_mounts: List[str] = None,
    git_diff_mount: str = None,
    local_files: List[str] = None,
    upload_local_git_diff: bool = None,
    archive_file_mount: str = None,
    root_volume_size: str = None,
    working_dir: str = None,
    output_dir: str = MOUNT_PATH_OUTPUT,
    worker_count: int = 1,
    framework_type: str = None,
    **kwargs,
) -> ResponseExperimentInfo:
    """Create experiment

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
        git_branch (str): override current git branch
        git_ref (str): override current git commit
        use_git_diff (bool): run experiment with uncommitted changes
        use_git_diff_untracked (bool): run with untracked changed (only valid if `use_git_diff` is set)
    """
    cluster = read_cluster(cluster_name)

    kernel_resource_spec = kernel_resource_spec_id = None
    if kernel_resource_spec_name:
        kernel_resource_spec_id = read_kernel_resource_spec(
            cluster.id,
            kernel_resource_spec_name,
        ).id
    else:
        kernel_resource_spec = _configure_custom_kernel_resource_spec(
            processor_type,
            cpu_limit,
            memory_limit,
            gpu_type,
            gpu_limit,
        )

    volume_mount_requests = _configure_volume_mount_requests(
        dataset_mounts=dataset_mounts,
        git_ref_mounts=git_ref_mounts,
        git_diff_mount=git_diff_mount,
        local_files=local_files,
        upload_local_git_diff=upload_local_git_diff,
        archive_file_mount=archive_file_mount,
        root_volume_size=root_volume_size,
        working_dir=working_dir,
        output_dir=output_dir,
        **kwargs,
    )

    if worker_count > 1:
        return vessl_api.distributed_experiment_create_api(
            organization_name=_get_organization_name(**kwargs),
            project_name=_get_project_name(**kwargs),
            distributed_experiment_create_api_payload=DistributedExperimentCreateAPIPayload(
                cluster_id=cluster.id,
                env_vars=[
                    OrmEnvVar(key, str(value))
                    for key, value in map(
                        lambda env_var: env_var.split("="), (env_vars or [])
                    )
                ],
                framework_type=framework_type,
                image_url=kernel_image_url,
                message=message,
                worker_replicas=worker_count,
                worker_resource_spec=kernel_resource_spec,
                worker_resource_spec_id=kernel_resource_spec_id,
                start_command=start_command,
                termination_protection=termination_protection,
                volumes=volume_mount_requests,
            ),
        )

    return vessl_api.experiment_create_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        experiment_create_api_payload=ExperimentCreateAPIPayload(
            cluster_id=cluster.id,
            env_vars=[
                OrmEnvVar(key, str(value))
                for key, value in map(
                    lambda env_var: env_var.split("="), (env_vars or [])
                )
            ],
            image_url=kernel_image_url,
            message=message,
            resource_spec=kernel_resource_spec,
            resource_spec_id=kernel_resource_spec_id,
            start_command=start_command,
            termination_protection=termination_protection,
            volumes=volume_mount_requests,
        ),
    )


def _create_orm_execution_environment():
    cpu_count = psutil.cpu_count()

    gpu_count = gpu_type = None
    try:
        pynvml.nvmlInit()
        gpu_count = pynvml.nvmlDeviceGetCount()
        gpu_type = pynvml.nvmlDeviceGetName(
            pynvml.nvmlDeviceGetHandleByIndex(0)
        ).decode("utf8")
    except pynvml.nvml.NVMLError:
        pass

    hostname = platform.node()
    os = platform.platform(aliased=True)
    python_version = platform.python_version()
    python_executable = sys.executable
    start_command = " ".join(sys.argv)
    vessl_version = __version__

    return OrmExecutionEnvironment(
        cpu_count=cpu_count,
        gpu_count=gpu_count,
        gpu_type=gpu_type,
        hostname=hostname,
        os=os,
        python_version=python_version,
        python_executable=python_executable,
        start_command=start_command,
        vessl_version=vessl_version,
    )


def create_local_experiment(message: str = None, **kwargs):
    return vessl_api.local_experiment_create_api(
        local_experiment_create_api_payload=LocalExperimentCreateAPIPayload(
            execution_environment=_create_orm_execution_environment(),
            message=message,
        ),
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
    )


def list_experiment_logs(
    experiment_name: str, tail: int = 200, worker_number: int = 0, after: int = 0, **kwargs
) -> List[InfluxdbWorkloadLog]:
    """List experiment logs

    Args:
        experiment_name (str): experiment name or number
        tail (int): number of lines to display from the end. Display all if -1.
        worker_number (int): override default worker number

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    if tail == -1:
        tail = None

    experiment = read_experiment(experiment_name, **kwargs)

    if experiment.is_distributed:
        str_worker_number = safe_cast(worker_number, str, "0")
        return vessl_api.distributed_experiment_logs_api(
            experiment=experiment_name,
            limit=tail,
            organization_name=_get_organization_name(**kwargs),
            project_name=_get_project_name(**kwargs),
            distributed_number=worker_number,
        ).logs[str_worker_number]

    return vessl_api.experiment_logs_api(
        experiment=experiment_name,
        limit=tail,
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        after=after,
    ).logs


def list_experiment_output_files(
    experiment_name: str,
    need_download_url: bool = False,
    recursive: bool = True,
    worker_number: int = 0,
    **kwargs,
) -> List[ResponseFileMetadata]:
    """List experiment output files

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    experiment = read_experiment(experiment_name, **kwargs)

    if experiment.is_local:
        volume_id = experiment.local_execution_spec.output_volume_id
        return list_volume_files(
            volume_id=volume_id,
            need_download_url=need_download_url,
            path="",
            recursive=recursive,
        )
    elif experiment.is_distributed:
        worker_replicas = experiment.distributed_spec.pytorch_spec.worker_replicas
        if worker_number >= worker_replicas:
            raise InvalidParamsError(
                "worker-number: should be less than {}".format(worker_replicas)
            )

        for volume_mount in experiment.volume_mounts_list[worker_number].mounts:
            if volume_mount.source_type == MOUNT_TYPE_OUTPUT:
                return list_volume_files(
                    volume_id=volume_mount.volume.volume_id,
                    need_download_url=need_download_url,
                    path="",
                    recursive=recursive,
                )
    else:
        for volume_mount in experiment.volume_mounts.mounts:
            if volume_mount.source_type == MOUNT_TYPE_OUTPUT:
                return list_volume_files(
                    volume_id=volume_mount.volume.volume_id,
                    need_download_url=need_download_url,
                    path="",
                    recursive=recursive,
                )

    logger.info("No output volume mounted")


def download_experiment_output_files(
    experiment_name: str,
    dest_path: str = os.path.join(os.getcwd(), "output"),
    worker_number: int = 0,
    **kwargs,
) -> None:
    """Download experiment output files

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    experiment = read_experiment(experiment_name, **kwargs)

    if experiment.is_local:
        volume_id = experiment.local_execution_spec.output_volume_id
        return copy_volume_file(
            source_volume_id=volume_id,
            source_path="",
            dest_volume_id=None,
            dest_path=dest_path,
            recursive=True,
        )
    if experiment.is_distributed:
        worker_replicas = experiment.distributed_spec.pytorch_spec.worker_replicas
        if worker_number >= worker_replicas:
            raise InvalidParamsError(
                "worker-number: should be less than {}".format(worker_replicas)
            )

        for volume_mount in experiment.volume_mounts_list[worker_number].mounts:
            if volume_mount.source_type == MOUNT_TYPE_OUTPUT:
                return copy_volume_file(
                    source_volume_id=volume_mount.volume.volume_id,
                    source_path="",
                    dest_volume_id=None,
                    dest_path=dest_path,
                    recursive=True,
                )
    else:
        for volume_mount in experiment.volume_mounts.mounts:
            if volume_mount.source_type == MOUNT_TYPE_OUTPUT:
                return copy_volume_file(
                    source_volume_id=volume_mount.volume.volume_id,
                    source_path="",
                    dest_volume_id=None,
                    dest_path=dest_path,
                    recursive=True,
                )

    logger.info("No output volume mounted")


def upload_experiment_output_files(experiment_name: str, path: str):
    experiment = read_experiment(experiment_name)
    if not experiment.is_local:
        raise InvalidExperimentError("Experiment cannot be Vessl-managed.")

    return copy_volume_file(
        source_volume_id=None,
        source_path=path,
        dest_volume_id=experiment.local_execution_spec.output_volume_id,
        dest_path="",
    )



import random
import string


def random_string(length=12, punctuations=False):
    letters = string.ascii_lowercase + string.digits
    if punctuations:
        letters += "!@#$^&"
    return random.SystemRandom().choice(string.ascii_lowercase) + "".join(
        random.SystemRandom().choice(letters) for _ in range(length - 1)
    )



# coding: utf-8

# flake8: noqa
"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from openapi_client.models.account_invitation_token_validate_response import AccountInvitationTokenValidateResponse
from openapi_client.models.account_ssh_key_list_response import AccountSSHKeyListResponse
from openapi_client.models.account_sign_in_cli_check_response import AccountSignInCliCheckResponse
from openapi_client.models.account_sign_in_cli_token_response import AccountSignInCliTokenResponse
from openapi_client.models.change_password_api_payload import ChangePasswordAPIPayload
from openapi_client.models.cli_experiment_freeze_dataset_version_api_payload import CliExperimentFreezeDatasetVersionAPIPayload
from openapi_client.models.cluster_cluster_list_response import ClusterClusterListResponse
from openapi_client.models.cluster_cluster_workload import ClusterClusterWorkload
from openapi_client.models.cluster_cluster_workload_list_response import ClusterClusterWorkloadListResponse
from openapi_client.models.cluster_custom_cluster_key_response import ClusterCustomClusterKeyResponse
from openapi_client.models.cluster_custom_cluster_list_response import ClusterCustomClusterListResponse
from openapi_client.models.cluster_custom_cluster_node_list_response import ClusterCustomClusterNodeListResponse
from openapi_client.models.cluster_managed_cluster_list_response import ClusterManagedClusterListResponse
from openapi_client.models.custom_cluster_add_api_payload import CustomClusterAddAPIPayload
from openapi_client.models.custom_cluster_update_api_payload import CustomClusterUpdateAPIPayload
from openapi_client.models.dashboard_chart_create_api_payload import DashboardChartCreateAPIPayload
from openapi_client.models.dashboard_chart_metric import DashboardChartMetric
from openapi_client.models.dashboard_chart_update_api_payload import DashboardChartUpdateAPIPayload
from openapi_client.models.dashboard_dashboard_chart_list_response import DashboardDashboardChartListResponse
from openapi_client.models.dashboard_dashboard_chart_metrics_response import DashboardDashboardChartMetricsResponse
from openapi_client.models.dashboard_dashboard_experiment_field_list_response import DashboardDashboardExperimentFieldListResponse
from openapi_client.models.dashboard_dashboard_experiment_field_update_response import DashboardDashboardExperimentFieldUpdateResponse
from openapi_client.models.dashboard_dashboard_experiment_field_value_list_response import DashboardDashboardExperimentFieldValueListResponse
from openapi_client.models.dashboard_dashboard_experiment_list_response import DashboardDashboardExperimentListResponse
from openapi_client.models.dashboard_dashboard_section_update_response import DashboardDashboardSectionUpdateResponse
from openapi_client.models.dashboard_experiment_field_update_api_payload import DashboardExperimentFieldUpdateAPIPayload
from openapi_client.models.dashboard_experiment_filter_create_api_payload import DashboardExperimentFilterCreateAPIPayload
from openapi_client.models.dashboard_experiment_filter_update_api_payload import DashboardExperimentFilterUpdateAPIPayload
from openapi_client.models.dashboard_experiment_sort_create_api_payload import DashboardExperimentSortCreateAPIPayload
from openapi_client.models.dashboard_experiment_sort_update_api_payload import DashboardExperimentSortUpdateAPIPayload
from openapi_client.models.dashboard_experiment_with_metrics import DashboardExperimentWithMetrics
from openapi_client.models.dashboard_field_request import DashboardFieldRequest
from openapi_client.models.dashboard_section_create_api_payload import DashboardSectionCreateAPIPayload
from openapi_client.models.dashboard_section_update_api_payload import DashboardSectionUpdateAPIPayload
from openapi_client.models.dataset_summary_update_api_payload import DatasetSummaryUpdateAPIPayload
from openapi_client.models.dataset_update_api_payload import DatasetUpdateAPIPayload
from openapi_client.models.dataset_version_create_api_payload import DatasetVersionCreateAPIPayload
from openapi_client.models.dataset_version_update_api_payload import DatasetVersionUpdateAPIPayload
from openapi_client.models.datasetversion_dataset_version_list_response import DatasetversionDatasetVersionListResponse
from openapi_client.models.distributed_experiment_create_api_payload import DistributedExperimentCreateAPIPayload
from openapi_client.models.distributed_experiment_distributed_experiment_logs_response import DistributedExperimentDistributedExperimentLogsResponse
from openapi_client.models.distributed_experiment_distributed_experiment_plots_response import DistributedExperimentDistributedExperimentPlotsResponse
from openapi_client.models.distributed_experiment_distributed_experiment_system_metrics_response import DistributedExperimentDistributedExperimentSystemMetricsResponse
from openapi_client.models.experiment_add_tag_api_payload import ExperimentAddTagAPIPayload
from openapi_client.models.experiment_bulk_delete_api_payload import ExperimentBulkDeleteAPIPayload
from openapi_client.models.experiment_bulk_terminate_api_payload import ExperimentBulkTerminateAPIPayload
from openapi_client.models.experiment_create_api_payload import ExperimentCreateAPIPayload
from openapi_client.models.experiment_experiment_add_tag_response import ExperimentExperimentAddTagResponse
from openapi_client.models.experiment_experiment_hide_plots_response import ExperimentExperimentHidePlotsResponse
from openapi_client.models.experiment_experiment_list_response import ExperimentExperimentListResponse
from openapi_client.models.experiment_experiment_logs_response import ExperimentExperimentLogsResponse
from openapi_client.models.experiment_experiment_metrics_update_response import ExperimentExperimentMetricsUpdateResponse
from openapi_client.models.experiment_experiment_parameter_list_response import ExperimentExperimentParameterListResponse
from openapi_client.models.experiment_experiment_plots_response import ExperimentExperimentPlotsResponse
from openapi_client.models.experiment_experiment_remove_tag_response import ExperimentExperimentRemoveTagResponse
from openapi_client.models.experiment_experiment_show_plots_response import ExperimentExperimentShowPlotsResponse
from openapi_client.models.experiment_experiment_status_idle_response import ExperimentExperimentStatusIdleResponse
from openapi_client.models.experiment_experiment_system_metrics_response import ExperimentExperimentSystemMetricsResponse
from openapi_client.models.experiment_hide_plots_api_payload import ExperimentHidePlotsAPIPayload
from openapi_client.models.experiment_metric_entry import ExperimentMetricEntry
from openapi_client.models.experiment_metrics_update_api_payload import ExperimentMetricsUpdateAPIPayload
from openapi_client.models.experiment_progress_update_api_payload import ExperimentProgressUpdateAPIPayload
from openapi_client.models.experiment_remove_tag_api_payload import ExperimentRemoveTagAPIPayload
from openapi_client.models.experiment_show_plots_api_payload import ExperimentShowPlotsAPIPayload
from openapi_client.models.experiment_update_api_payload import ExperimentUpdateAPIPayload
from openapi_client.models.external_git_hub_branch_list_response import ExternalGitHubBranchListResponse
from openapi_client.models.external_git_hub_commit_list_response import ExternalGitHubCommitListResponse
from openapi_client.models.external_git_hub_config_response import ExternalGitHubConfigResponse
from openapi_client.models.external_google_config_response import ExternalGoogleConfigResponse
from openapi_client.models.external_slack_config_response import ExternalSlackConfigResponse
from openapi_client.models.gs_dataset_create_api_payload import GSDatasetCreateAPIPayload
from openapi_client.models.github_code_of_conduct import GithubCodeOfConduct
from openapi_client.models.github_license import GithubLicense
from openapi_client.models.github_match import GithubMatch
from openapi_client.models.github_organization import GithubOrganization
from openapi_client.models.github_plan import GithubPlan
from openapi_client.models.github_repository import GithubRepository
from openapi_client.models.github_text_match import GithubTextMatch
from openapi_client.models.github_user import GithubUser
from openapi_client.models.influxdb_current_system_metric import InfluxdbCurrentSystemMetric
from openapi_client.models.influxdb_experiment_plot_metric import InfluxdbExperimentPlotMetric
from openapi_client.models.influxdb_metric_legacy import InfluxdbMetricLegacy
from openapi_client.models.influxdb_sweep_log import InfluxdbSweepLog
from openapi_client.models.influxdb_system_metric import InfluxdbSystemMetric
from openapi_client.models.influxdb_system_metric_list import InfluxdbSystemMetricList
from openapi_client.models.influxdb_workload_log import InfluxdbWorkloadLog
from openapi_client.models.invitation_token_validate_api_payload import InvitationTokenValidateAPIPayload
from openapi_client.models.kernel_kernel_image_list_response import KernelKernelImageListResponse
from openapi_client.models.kernel_kernel_resource_spec_list_response import KernelKernelResourceSpecListResponse
from openapi_client.models.kernel_resource_spec_create_api_payload import KernelResourceSpecCreateAPIPayload
from openapi_client.models.kernel_resource_spec_update_api_payload import KernelResourceSpecUpdateAPIPayload
from openapi_client.models.local_dataset_create_api_payload import LocalDatasetCreateAPIPayload
from openapi_client.models.local_experiment_create_api_payload import LocalExperimentCreateAPIPayload
from openapi_client.models.local_experiment_finish_api_payload import LocalExperimentFinishAPIPayload
from openapi_client.models.model_add_tag_api_payload import ModelAddTagAPIPayload
from openapi_client.models.model_create_api_payload import ModelCreateAPIPayload
from openapi_client.models.model_model_add_tag_response import ModelModelAddTagResponse
from openapi_client.models.model_model_remove_tag_response import ModelModelRemoveTagResponse
from openapi_client.models.model_remove_tag_api_payload import ModelRemoveTagAPIPayload
from openapi_client.models.model_update_api_payload import ModelUpdateAPIPayload
from openapi_client.models.organization_aws_credentials_add_api_payload import OrganizationAWSCredentialsAddAPIPayload
from openapi_client.models.organization_create_api_payload import OrganizationCreateAPIPayload
from openapi_client.models.organization_docker_credentials_add_api_payload import OrganizationDockerCredentialsAddAPIPayload
from openapi_client.models.organization_git_hub_credentials_add_api_payload import OrganizationGitHubCredentialsAddAPIPayload
from openapi_client.models.organization_member_add_api_payload import OrganizationMemberAddAPIPayload
from openapi_client.models.organization_notification_config_email_add_api_payload import OrganizationNotificationConfigEmailAddAPIPayload
from openapi_client.models.organization_org_quota import OrganizationOrgQuota
from openapi_client.models.organization_organization_billing_history_filter_list_response import OrganizationOrganizationBillingHistoryFilterListResponse
from openapi_client.models.organization_organization_billing_history_list_response import OrganizationOrganizationBillingHistoryListResponse
from openapi_client.models.organization_organization_billing_summary_response import OrganizationOrganizationBillingSummaryResponse
from openapi_client.models.organization_organization_billing_total_response import OrganizationOrganizationBillingTotalResponse
from openapi_client.models.organization_organization_git_repository_list_response import OrganizationOrganizationGitRepositoryListResponse
from openapi_client.models.organization_organization_list_response import OrganizationOrganizationListResponse
from openapi_client.models.organization_organization_me_response import OrganizationOrganizationMeResponse
from openapi_client.models.organization_organization_member_list_response import OrganizationOrganizationMemberListResponse
from openapi_client.models.organization_organization_quota_list_response import OrganizationOrganizationQuotaListResponse
from openapi_client.models.organization_organization_quota_me_response import OrganizationOrganizationQuotaMeResponse
from openapi_client.models.organization_quota_upsert_api_payload import OrganizationQuotaUpsertAPIPayload
from openapi_client.models.organization_update_api_payload import OrganizationUpdateAPIPayload
from openapi_client.models.organization_user_quota import OrganizationUserQuota
from openapi_client.models.orm_access_token import OrmAccessToken
from openapi_client.models.orm_access_token_edges import OrmAccessTokenEdges
from openapi_client.models.orm_billing_history import OrmBillingHistory
from openapi_client.models.orm_billing_history_edges import OrmBillingHistoryEdges
from openapi_client.models.orm_distributed_py_torch_spec import OrmDistributedPyTorchSpec
from openapi_client.models.orm_distributed_spec import OrmDistributedSpec
from openapi_client.models.orm_duration_value import OrmDurationValue
from openapi_client.models.orm_early_stopping_setting import OrmEarlyStoppingSetting
from openapi_client.models.orm_early_stopping_spec import OrmEarlyStoppingSpec
from openapi_client.models.orm_env_var import OrmEnvVar
from openapi_client.models.orm_env_vars import OrmEnvVars
from openapi_client.models.orm_execution_environment import OrmExecutionEnvironment
from openapi_client.models.orm_experiment import OrmExperiment
from openapi_client.models.orm_experiment_edges import OrmExperimentEdges
from openapi_client.models.orm_experiment_filter_values import OrmExperimentFilterValues
from openapi_client.models.orm_experiment_metrics_summary import OrmExperimentMetricsSummary
from openapi_client.models.orm_file_metadata import OrmFileMetadata
from openapi_client.models.orm_kernel_cluster import OrmKernelCluster
from openapi_client.models.orm_kernel_cluster_config import OrmKernelClusterConfig
from openapi_client.models.orm_kernel_cluster_config_ingress import OrmKernelClusterConfigIngress
from openapi_client.models.orm_kernel_cluster_config_nodes import OrmKernelClusterConfigNodes
from openapi_client.models.orm_kernel_cluster_config_service import OrmKernelClusterConfigService
from openapi_client.models.orm_kernel_cluster_config_storage_class import OrmKernelClusterConfigStorageClass
from openapi_client.models.orm_kernel_cluster_edges import OrmKernelClusterEdges
from openapi_client.models.orm_kernel_cluster_node import OrmKernelClusterNode
from openapi_client.models.orm_kernel_cluster_node_edges import OrmKernelClusterNodeEdges
from openapi_client.models.orm_kernel_cluster_select_policies import OrmKernelClusterSelectPolicies
from openapi_client.models.orm_kernel_cluster_select_policy import OrmKernelClusterSelectPolicy
from openapi_client.models.orm_kernel_cluster_storage import OrmKernelClusterStorage
from openapi_client.models.orm_kernel_cluster_storage_edges import OrmKernelClusterStorageEdges
from openapi_client.models.orm_kernel_image import OrmKernelImage
from openapi_client.models.orm_kernel_image_edges import OrmKernelImageEdges
from openapi_client.models.orm_kernel_resource_spec import OrmKernelResourceSpec
from openapi_client.models.orm_kernel_resource_spec_edges import OrmKernelResourceSpecEdges
from openapi_client.models.orm_kernel_resource_spec_field import OrmKernelResourceSpecField
from openapi_client.models.orm_kernel_resource_spec_node_selector import OrmKernelResourceSpecNodeSelector
from openapi_client.models.orm_local_execution_spec import OrmLocalExecutionSpec
from openapi_client.models.orm_local_volume_config import OrmLocalVolumeConfig
from openapi_client.models.orm_local_volume_config_flex_volume import OrmLocalVolumeConfigFlexVolume
from openapi_client.models.orm_local_volume_config_flex_volume_options import OrmLocalVolumeConfigFlexVolumeOptions
from openapi_client.models.orm_local_volume_config_host_path import OrmLocalVolumeConfigHostPath
from openapi_client.models.orm_local_volume_config_nfs import OrmLocalVolumeConfigNFS
from openapi_client.models.orm_model import OrmModel
from openapi_client.models.orm_model_edges import OrmModelEdges
from openapi_client.models.orm_notification_config import OrmNotificationConfig
from openapi_client.models.orm_notification_config_edges import OrmNotificationConfigEdges
from openapi_client.models.orm_objective_step_median import OrmObjectiveStepMedian
from openapi_client.models.orm_organization import OrmOrganization
from openapi_client.models.orm_organization_credentials import OrmOrganizationCredentials
from openapi_client.models.orm_organization_credentials_edges import OrmOrganizationCredentialsEdges
from openapi_client.models.orm_organization_edges import OrmOrganizationEdges
from openapi_client.models.orm_organization_kernel_cluster import OrmOrganizationKernelCluster
from openapi_client.models.orm_organization_kernel_cluster_edges import OrmOrganizationKernelClusterEdges
from openapi_client.models.orm_page_info_with_count import OrmPageInfoWithCount
from openapi_client.models.orm_parameter import OrmParameter
from openapi_client.models.orm_project import OrmProject
from openapi_client.models.orm_project_edges import OrmProjectEdges
from openapi_client.models.orm_project_repository import OrmProjectRepository
from openapi_client.models.orm_project_repository_edges import OrmProjectRepositoryEdges
from openapi_client.models.orm_quota import OrmQuota
from openapi_client.models.orm_quota_edges import OrmQuotaEdges
from openapi_client.models.orm_range import OrmRange
from openapi_client.models.orm_step_median import OrmStepMedian
from openapi_client.models.orm_storage import OrmStorage
from openapi_client.models.orm_storage_edges import OrmStorageEdges
from openapi_client.models.orm_suggestion_histories import OrmSuggestionHistories
from openapi_client.models.orm_suggestion_history import OrmSuggestionHistory
from openapi_client.models.orm_suggestion_history_parameter import OrmSuggestionHistoryParameter
from openapi_client.models.orm_sweep import OrmSweep
from openapi_client.models.orm_sweep_edges import OrmSweepEdges
from openapi_client.models.orm_sweep_history import OrmSweepHistory
from openapi_client.models.orm_sweep_history_edges import OrmSweepHistoryEdges
from openapi_client.models.orm_sweep_objective import OrmSweepObjective
from openapi_client.models.orm_sweep_search_space import OrmSweepSearchSpace
from openapi_client.models.orm_tag import OrmTag
from openapi_client.models.orm_tag_edges import OrmTagEdges
from openapi_client.models.orm_user import OrmUser
from openapi_client.models.orm_user_edges import OrmUserEdges
from openapi_client.models.orm_user_organization import OrmUserOrganization
from openapi_client.models.orm_user_organization_edges import OrmUserOrganizationEdges
from openapi_client.models.orm_volume import OrmVolume
from openapi_client.models.orm_volume_edges import OrmVolumeEdges
from openapi_client.models.orm_volume_source_archive_file import OrmVolumeSourceArchiveFile
from openapi_client.models.orm_volume_source_code import OrmVolumeSourceCode
from openapi_client.models.orm_volume_source_dataset import OrmVolumeSourceDataset
from openapi_client.models.orm_volume_source_dataset_version import OrmVolumeSourceDatasetVersion
from openapi_client.models.orm_volume_source_volume import OrmVolumeSourceVolume
from openapi_client.models.orm_workload import OrmWorkload
from openapi_client.models.orm_workload_edges import OrmWorkloadEdges
from openapi_client.models.orm_workload_endpoint import OrmWorkloadEndpoint
from openapi_client.models.orm_workload_history import OrmWorkloadHistory
from openapi_client.models.orm_workload_history_edges import OrmWorkloadHistoryEdges
from openapi_client.models.orm_workspace import OrmWorkspace
from openapi_client.models.orm_workspace_edges import OrmWorkspaceEdges
from openapi_client.models.orm_workspace_port import OrmWorkspacePort
from openapi_client.models.orm_workspace_ports import OrmWorkspacePorts
from openapi_client.models.project_add_project_repository_api_payload import ProjectAddProjectRepositoryAPIPayload
from openapi_client.models.project_create_api_payload import ProjectCreateAPIPayload
from openapi_client.models.project_project_branch_list_response import ProjectProjectBranchListResponse
from openapi_client.models.project_project_commit_list_response import ProjectProjectCommitListResponse
from openapi_client.models.project_project_repository_branch_list_response import ProjectProjectRepositoryBranchListResponse
from openapi_client.models.project_project_repository_commit_list_response import ProjectProjectRepositoryCommitListResponse
from openapi_client.models.project_remove_project_repository_api_payload import ProjectRemoveProjectRepositoryAPIPayload
from openapi_client.models.project_tag_create_api_payload import ProjectTagCreateAPIPayload
from openapi_client.models.project_update_api_payload import ProjectUpdateAPIPayload
from openapi_client.models.proto_branch import ProtoBranch
from openapi_client.models.proto_commit import ProtoCommit
from openapi_client.models.proto_commit_author import ProtoCommitAuthor
from openapi_client.models.proto_project_repository_request import ProtoProjectRepositoryRequest
from openapi_client.models.proto_tag import ProtoTag
from openapi_client.models.proto_volume_mount_request import ProtoVolumeMountRequest
from openapi_client.models.proto_volume_mount_request_source_archive_file import ProtoVolumeMountRequestSourceArchiveFile
from openapi_client.models.proto_volume_mount_request_source_code import ProtoVolumeMountRequestSourceCode
from openapi_client.models.proto_volume_mount_request_source_dataset import ProtoVolumeMountRequestSourceDataset
from openapi_client.models.proto_volume_mount_request_source_dataset_version import ProtoVolumeMountRequestSourceDatasetVersion
from openapi_client.models.proto_volume_mount_request_source_volume import ProtoVolumeMountRequestSourceVolume
from openapi_client.models.proto_volume_mount_requests import ProtoVolumeMountRequests
from openapi_client.models.region_region_list_response import RegionRegionListResponse
from openapi_client.models.response_billing_history_info import ResponseBillingHistoryInfo
from openapi_client.models.response_code_repository import ResponseCodeRepository
from openapi_client.models.response_daily_billing import ResponseDailyBilling
from openapi_client.models.response_dashboard_chart import ResponseDashboardChart
from openapi_client.models.response_dashboard_chart_field import ResponseDashboardChartField
from openapi_client.models.response_dashboard_chart_section import ResponseDashboardChartSection
from openapi_client.models.response_dashboard_experiment_field import ResponseDashboardExperimentField
from openapi_client.models.response_dashboard_experiment_filter_response import ResponseDashboardExperimentFilterResponse
from openapi_client.models.response_dashboard_experiment_filter_values import ResponseDashboardExperimentFilterValues
from openapi_client.models.response_dashboard_experiment_sort_response import ResponseDashboardExperimentSortResponse
from openapi_client.models.response_dashboard_response import ResponseDashboardResponse
from openapi_client.models.response_dataset_info import ResponseDatasetInfo
from openapi_client.models.response_dataset_info_detail import ResponseDatasetInfoDetail
from openapi_client.models.response_dataset_info_list import ResponseDatasetInfoList
from openapi_client.models.response_dataset_source import ResponseDatasetSource
from openapi_client.models.response_dataset_summary import ResponseDatasetSummary
from openapi_client.models.response_dataset_version_info import ResponseDatasetVersionInfo
from openapi_client.models.response_experiment_info import ResponseExperimentInfo
from openapi_client.models.response_experiment_list_response import ResponseExperimentListResponse
from openapi_client.models.response_experiment_plot_file import ResponseExperimentPlotFile
from openapi_client.models.response_field_object_value import ResponseFieldObjectValue
from openapi_client.models.response_file_metadata import ResponseFileMetadata
from openapi_client.models.response_kernel_cluster import ResponseKernelCluster
from openapi_client.models.response_kernel_cluster_info import ResponseKernelClusterInfo
from openapi_client.models.response_kernel_cluster_node import ResponseKernelClusterNode
from openapi_client.models.response_kernel_cluster_node_info import ResponseKernelClusterNodeInfo
from openapi_client.models.response_kernel_cluster_system_metrics import ResponseKernelClusterSystemMetrics
from openapi_client.models.response_kernel_image import ResponseKernelImage
from openapi_client.models.response_kernel_resource_spec import ResponseKernelResourceSpec
from openapi_client.models.response_kubernetes_resource_info import ResponseKubernetesResourceInfo
from openapi_client.models.response_model_detail import ResponseModelDetail
from openapi_client.models.response_model_detail_list import ResponseModelDetailList
from openapi_client.models.response_my_user import ResponseMyUser
from openapi_client.models.response_organization import ResponseOrganization
from openapi_client.models.response_organization_credentials_info import ResponseOrganizationCredentialsInfo
from openapi_client.models.response_organization_credentials_info_list import ResponseOrganizationCredentialsInfoList
from openapi_client.models.response_organization_info import ResponseOrganizationInfo
from openapi_client.models.response_organization_member import ResponseOrganizationMember
from openapi_client.models.response_organization_notification_config import ResponseOrganizationNotificationConfig
from openapi_client.models.response_parameter_info import ResponseParameterInfo
from openapi_client.models.response_project import ResponseProject
from openapi_client.models.response_project_billing import ResponseProjectBilling
from openapi_client.models.response_project_experiment_summary import ResponseProjectExperimentSummary
from openapi_client.models.response_project_info import ResponseProjectInfo
from openapi_client.models.response_project_list_response import ResponseProjectListResponse
from openapi_client.models.response_project_repository import ResponseProjectRepository
from openapi_client.models.response_region import ResponseRegion
from openapi_client.models.response_ssh_key_info import ResponseSSHKeyInfo
from openapi_client.models.response_simple_experiment import ResponseSimpleExperiment
from openapi_client.models.response_simple_project import ResponseSimpleProject
from openapi_client.models.response_simple_sweep import ResponseSimpleSweep
from openapi_client.models.response_simple_user import ResponseSimpleUser
from openapi_client.models.response_simple_workspace import ResponseSimpleWorkspace
from openapi_client.models.response_sweep_experiment_info import ResponseSweepExperimentInfo
from openapi_client.models.response_sweep_experiment_summary import ResponseSweepExperimentSummary
from openapi_client.models.response_sweep_history_info import ResponseSweepHistoryInfo
from openapi_client.models.response_sweep_info import ResponseSweepInfo
from openapi_client.models.response_sweep_list_response import ResponseSweepListResponse
from openapi_client.models.response_tag_response import ResponseTagResponse
from openapi_client.models.response_user import ResponseUser
from openapi_client.models.response_user_billing import ResponseUserBilling
from openapi_client.models.response_user_with_token_response import ResponseUserWithTokenResponse
from openapi_client.models.response_volume import ResponseVolume
from openapi_client.models.response_volume_mount_info import ResponseVolumeMountInfo
from openapi_client.models.response_volume_mount_infos import ResponseVolumeMountInfos
from openapi_client.models.response_volume_source_dataset import ResponseVolumeSourceDataset
from openapi_client.models.response_volume_source_dataset_version import ResponseVolumeSourceDatasetVersion
from openapi_client.models.response_volume_source_volume import ResponseVolumeSourceVolume
from openapi_client.models.response_workload_endpoints import ResponseWorkloadEndpoints
from openapi_client.models.response_workload_history_info import ResponseWorkloadHistoryInfo
from openapi_client.models.response_workload_status_info import ResponseWorkloadStatusInfo
from openapi_client.models.response_workspace_detail import ResponseWorkspaceDetail
from openapi_client.models.response_workspace_list import ResponseWorkspaceList
from openapi_client.models.s3_dataset_create_api_payload import S3DatasetCreateAPIPayload
from openapi_client.models.ssh_key_create_api_payload import SSHKeyCreateAPIPayload
from openapi_client.models.savvi_hub_dataset_create_api_payload import SavviHubDatasetCreateAPIPayload
from openapi_client.models.sign_in_api_payload import SignInAPIPayload
from openapi_client.models.sign_in_cli_confirm_api_payload import SignInCliConfirmAPIPayload
from openapi_client.models.sign_in_git_hub_api_payload import SignInGitHubAPIPayload
from openapi_client.models.sign_in_google_api_payload import SignInGoogleAPIPayload
from openapi_client.models.sign_up_api_payload import SignUpAPIPayload
from openapi_client.models.sign_up_validate_api_payload import SignUpValidateAPIPayload
from openapi_client.models.slack_authorize_api_payload import SlackAuthorizeAPIPayload
from openapi_client.models.storage_federation_token import StorageFederationToken
from openapi_client.models.storage_file_action_url_info import StorageFileActionURLInfo
from openapi_client.models.sweep_create_api_payload import SweepCreateAPIPayload
from openapi_client.models.sweep_sweep_list_response import SweepSweepListResponse
from openapi_client.models.sweep_sweep_logs_response import SweepSweepLogsResponse
from openapi_client.models.sweep_sweep_plots_response import SweepSweepPlotsResponse
from openapi_client.models.sweep_sweep_system_metrics_response import SweepSweepSystemMetricsResponse
from openapi_client.models.tag_experiment_tag_list_response import TagExperimentTagListResponse
from openapi_client.models.tag_update_api_payload import TagUpdateAPIPayload
from openapi_client.models.v1_node_selector_requirement import V1NodeSelectorRequirement
from openapi_client.models.v1_toleration import V1Toleration
from openapi_client.models.verify_email_api_payload import VerifyEmailAPIPayload
from openapi_client.models.volume_file_copy_api_payload import VolumeFileCopyAPIPayload
from openapi_client.models.volume_file_create_api_payload import VolumeFileCreateAPIPayload
from openapi_client.models.volume_volume_file_copy_response import VolumeVolumeFileCopyResponse
from openapi_client.models.volume_volume_file_delete_response import VolumeVolumeFileDeleteResponse
from openapi_client.models.volume_volume_file_list_response import VolumeVolumeFileListResponse
from openapi_client.models.workspace_cli_workspace_backup_response import WorkspaceCliWorkspaceBackupResponse
from openapi_client.models.workspace_create_api_payload import WorkspaceCreateAPIPayload
from openapi_client.models.workspace_update_api_payload import WorkspaceUpdateAPIPayload
from openapi_client.models.workspace_workspace_list_response import WorkspaceWorkspaceListResponse
from openapi_client.models.workspace_workspace_logs_response import WorkspaceWorkspaceLogsResponse
from openapi_client.models.workspace_workspace_status_running_response import WorkspaceWorkspaceStatusRunningResponse
from openapi_client.models.workspace_workspace_system_metrics_response import WorkspaceWorkspaceSystemMetricsResponse



import os
from urllib.parse import urlparse

import click

from vessl.cli._base import VesslGroup, vessl_argument, vessl_option
from vessl.cli._util import generic_prompter, print_volume_files
from vessl.util.exception import InvalidVolumeFileError
from vessl.volume import copy_volume_file, delete_volume_file, list_volume_files


@click.command(name="volume", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_argument(
    "id",
    type=click.INT,
    required=True,
    prompter=generic_prompter("Volume ID", click.INT),
)
@click.option("-p", "--path", type=click.Path(), default="", help="Defaults to root.")
@click.option("-r", "--recursive", is_flag=True)
def list(id: int, path: str, recursive: bool):
    files = list_volume_files(
        volume_id=id,
        path=path,
        need_download_url=False,
        recursive=recursive,
    )
    print_volume_files(files)


@cli.vessl_command()
@vessl_argument(
    "id",
    type=click.INT,
    required=True,
    prompter=generic_prompter("Volume ID", click.INT),
)
@click.option("-r", "--recursive", is_flag=True, help="Required if directory.")
def delete(id: int, recursive: bool):
    files = delete_volume_file(volume_id=id, recursive=recursive)
    print(f"Deleted {id} ({len([x for x in files if not x.is_dir])} file(s)).")


@cli.vessl_command()
@vessl_option(
    "--source-id",
    type=click.INT,
    prompter=generic_prompter("Source volume ID", click.INT),
    help="If not specified, source is assumed to be local.",
)
@vessl_option(
    "--source-path",
    type=click.Path(),
    required=True,
    prompter=generic_prompter("Source path"),
)
@vessl_option(
    "--dest-id",
    type=click.INT,
    prompter=generic_prompter("Destination volume ID", click.INT),
    help="If not specified, destination is assumed to be local.",
)
@vessl_option(
    "--dest-path",
    type=click.Path(),
    required=True,
    prompter=generic_prompter("Destination path"),
)
@click.option("-r", "--recursive", is_flag=True, help="Required if directory.")
def copy(
    source_id: int, source_path: str, dest_id: int, dest_path: str, recursive: bool
):
    copy_volume_file(
        source_volume_id=source_id,
        source_path=source_path,
        dest_volume_id=dest_id,
        dest_path=dest_path,
        recursive=recursive,
    )


def parse_volume_url(volume_url: str, raise_if_not_exists: bool = False):
    if volume_url.startswith("sv://"):
        u = urlparse(volume_url)
        volume_id = int(u.netloc.split(":", 1)[0])
        path = u.path.lstrip("/")
        return volume_id, path

    path = os.path.abspath(volume_url)
    if not os.path.exists(path) and raise_if_not_exists:
        raise InvalidVolumeFileError(f"No such file: {volume_url}")

    return None, path


@cli.vessl_command(hidden=True)
@click.argument(
    "source",
    type=click.STRING,
    required=True,
)
@click.argument(
    "dest",
    type=click.STRING,
    required=True,
)
@click.option("-r", "--recursive", is_flag=True)
def cp(source: str, dest: str, recursive: bool):
    """Volume copy method used in backend"""

    source_id, source_path = parse_volume_url(source, raise_if_not_exists=True)
    dest_id, dest_path = parse_volume_url(dest)

    copy_volume_file(
        source_volume_id=source_id,
        source_path=source_path,
        dest_volume_id=dest_id,
        dest_path=dest_path,
        recursive=recursive,
    )



from typing import List

from openapi_client.models import (
    ModelCreateAPIPayload,
    ModelUpdateAPIPayload,
    ResponseModelDetail,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.project import _get_project_name


def read_model(version: int, **kwargs) -> ResponseModelDetail:
    """Read model

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    return vessl_api.model_read_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        version=version,
    )


def list_models(**kwargs) -> List[ResponseModelDetail]:
    """List models

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    return vessl_api.model_list_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
    ).results


def create_model(
    experiment_id: int, description: str = None, tags: List[str] = None, **kwargs
) -> ResponseModelDetail:
    """Create model

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    return vessl_api.model_create_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        model_create_api_payload=ModelCreateAPIPayload(
            experiment_id=experiment_id,
            description=description,
            tag_names=tags,
        ),
    )


def update_model(version: int, description: str, **kwargs) -> ResponseModelDetail:
    """Update model

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    return vessl_api.model_update_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        version=version,
        model_update_api_payload=ModelUpdateAPIPayload(description=description),
    )


def delete_model(version: int, **kwargs) -> object:
    """Delete model

    Keyword args:
        organization_name (str): override default organization
        project_name (str): override default project
    """
    return vessl_api.model_delete_api(
        organization_name=_get_organization_name(**kwargs),
        project_name=_get_project_name(**kwargs),
        version=version,
    )



from test.conftest import USE_MOCK
from test.util import random_string

import pytest
from mock import Mock, patch

import vessl
from vessl.util.exception import InvalidProjectError


@pytest.mark.skipif(USE_MOCK, reason="Does not run if mocking is used.")
class TestProject:
    project_name = random_string()

    @pytest.mark.order(index=1)
    def test_create_project(self):
        vessl.create_project(self.project_name)

    def test_read_project(self):
        vessl.read_project(self.project_name)

    def test_list_projects(self):
        vessl.list_projects()


def test_get_project_name():
    with patch.object(vessl.vessl_api, "project", None):
        with pytest.raises(InvalidProjectError):
            vessl.project._get_project_name()

    project_name = "proj"
    assert vessl.project._get_project_name(project_name=project_name) == project_name

    project = Mock()
    project.name = project_name
    with patch.object(vessl.vessl_api, "project", project):
        assert vessl.project._get_project_name() == project_name


def test_get_project():
    with patch.object(vessl.vessl_api, "project", None):
        with pytest.raises(InvalidProjectError):
            vessl.project._get_project()

    project_name = "proj"

    project = Mock()
    project.name = project_name
    with patch.object(vessl.project, "read_project", return_value=project):
        assert (
            vessl.project._get_project(project_name=project_name).name == project_name
        )

    with patch.object(vessl.vessl_api, "project", project):
        assert vessl.project._get_project().name == project_name



import base64
import json
import os

import boto3
import requests
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from vessl.util import logger


class UploadableFileObject:
    def __init__(self, url, base_path, path):
        self.url = url
        self.base_path = base_path
        self.full_path = os.path.join(base_path, path)
        self.path = path

    def read_in_chunks(filename, chunk_size=65535, chunks=-1, callback=None):
        """Lazy function (generator) to read a file piece by piece."""
        with open(filename, "rb") as f:
            while chunks:
                data = f.read(chunk_size)
                if not data:
                    break
                yield data

                if callback:
                    callback(data)
                chunks -= 1

    def upload_chunks(self, *, callback=None):
        return self.read_in_chunks(self.full_path, callback=callback)

    def upload_hooks(self, *, callback=None):
        def fn(resp, **kwargs):
            if resp.status_code != 200:
                logger.warning(
                    f"Upload for {resp.request.url} failed. Detail: {resp.data}"
                )

        return {
            "response": fn,
        }

    def upload(self, session=requests.Session()):
        file_size = os.path.getsize(self.full_path)

        # send empty data when file is empty
        if os.stat(self.full_path).st_size == 0:
            future = session.put(
                self.url,
                data="",
                headers={"content-type": "application/octet-stream"},
                hooks=self.upload_hooks(),
            )
            return future

        with open(self.full_path, "rb") as f:
            with tqdm(
                total=file_size,
                desc=self.path,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as t:
                wrapped_file = CallbackIOWrapper(t.update, f, "read")
                requests.put(self.url, data=wrapped_file)
        return


class UploadableS3Object:
    def __init__(self, local_path: str, bucket, key, token, verbose=False):
        self.path = local_path
        self.bucket = bucket
        self.key = key
        self.token = token

        self.s3_client = self._get_s3_client_from_token()

        if verbose:
            boto3.set_stream_logger(name="botocore")

    def upload(self):
        self.s3_client.upload_file(self.path, self.bucket, self.key)

    def _get_s3_client_from_token(self):
        credentials = base64.b64decode(self.token).decode()
        credentials_dict = json.loads(credentials)
        return boto3.client(
            "s3",
            aws_access_key_id=credentials_dict["AccessKeyId"],
            aws_secret_access_key=credentials_dict["SecretAccessKey"],
            aws_session_token=credentials_dict["SessionToken"],
        )


class DownloadableFileObject:
    def __init__(self, url, base_path, path, size=None):
        self.url = url
        self.full_path = os.path.join(base_path, path)
        self.size = size

    def download_hooks(self, *, callback=None):
        def fn(resp, **kwargs):
            dirname = os.path.dirname(self.full_path)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            with open(self.full_path, "wb") as f:
                for chunk in resp.iter_content(chunk_size=8192):
                    f.write(chunk)
                    if callback:
                        callback(chunk)
            logger.debug(f"Downloaded {self.full_path}.")

        return {"response": fn}

    def download(self, session=requests.Session(), progressable=None):
        progress_callback = None
        if progressable and self.size:
            progress = progressable(length=self.size, label=self.full_path)
            progress_callback = lambda data: progress.update(len(data))

        future = session.get(
            self.url, stream=True, hooks=self.download_hooks(callback=progress_callback)
        )
        return future



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmExperiment(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'distributed_spec': 'dict(str, object)',
        'edges': 'OrmExperimentEdges',
        'estimated_end_dt': 'datetime',
        'exit_code': 'int',
        'experiment_created_by': 'int',
        'experiment_plot_volume': 'int',
        'experiment_project': 'int',
        'experiment_sweep': 'int',
        'id': 'int',
        'immutable_slug': 'str',
        'is_distributed': 'bool',
        'is_local': 'bool',
        'last_alive': 'datetime',
        'local_execution_spec': 'dict(str, object)',
        'message': 'str',
        'metrics_summary': 'dict(str, object)',
        'name': 'str',
        'number': 'int',
        'progress_percent': 'float',
        'show_plots': 'bool',
        'status': 'str',
        'status_last_updated': 'datetime',
        'status_reason': 'str',
        'termination_protection': 'bool',
        'updated_dt': 'datetime'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'distributed_spec': 'distributed_spec',
        'edges': 'edges',
        'estimated_end_dt': 'estimated_end_dt',
        'exit_code': 'exit_code',
        'experiment_created_by': 'experiment_created_by',
        'experiment_plot_volume': 'experiment_plot_volume',
        'experiment_project': 'experiment_project',
        'experiment_sweep': 'experiment_sweep',
        'id': 'id',
        'immutable_slug': 'immutable_slug',
        'is_distributed': 'is_distributed',
        'is_local': 'is_local',
        'last_alive': 'last_alive',
        'local_execution_spec': 'local_execution_spec',
        'message': 'message',
        'metrics_summary': 'metrics_summary',
        'name': 'name',
        'number': 'number',
        'progress_percent': 'progress_percent',
        'show_plots': 'show_plots',
        'status': 'status',
        'status_last_updated': 'status_last_updated',
        'status_reason': 'status_reason',
        'termination_protection': 'termination_protection',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, created_dt=None, distributed_spec=None, edges=None, estimated_end_dt=None, exit_code=None, experiment_created_by=None, experiment_plot_volume=None, experiment_project=None, experiment_sweep=None, id=None, immutable_slug=None, is_distributed=None, is_local=None, last_alive=None, local_execution_spec=None, message=None, metrics_summary=None, name=None, number=None, progress_percent=None, show_plots=None, status=None, status_last_updated=None, status_reason=None, termination_protection=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """OrmExperiment - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._distributed_spec = None
        self._edges = None
        self._estimated_end_dt = None
        self._exit_code = None
        self._experiment_created_by = None
        self._experiment_plot_volume = None
        self._experiment_project = None
        self._experiment_sweep = None
        self._id = None
        self._immutable_slug = None
        self._is_distributed = None
        self._is_local = None
        self._last_alive = None
        self._local_execution_spec = None
        self._message = None
        self._metrics_summary = None
        self._name = None
        self._number = None
        self._progress_percent = None
        self._show_plots = None
        self._status = None
        self._status_last_updated = None
        self._status_reason = None
        self._termination_protection = None
        self._updated_dt = None
        self.discriminator = None

        if created_dt is not None:
            self.created_dt = created_dt
        if distributed_spec is not None:
            self.distributed_spec = distributed_spec
        if edges is not None:
            self.edges = edges
        self.estimated_end_dt = estimated_end_dt
        self.exit_code = exit_code
        if experiment_created_by is not None:
            self.experiment_created_by = experiment_created_by
        if experiment_plot_volume is not None:
            self.experiment_plot_volume = experiment_plot_volume
        if experiment_project is not None:
            self.experiment_project = experiment_project
        self.experiment_sweep = experiment_sweep
        if id is not None:
            self.id = id
        if immutable_slug is not None:
            self.immutable_slug = immutable_slug
        if is_distributed is not None:
            self.is_distributed = is_distributed
        if is_local is not None:
            self.is_local = is_local
        if last_alive is not None:
            self.last_alive = last_alive
        if local_execution_spec is not None:
            self.local_execution_spec = local_execution_spec
        if message is not None:
            self.message = message
        if metrics_summary is not None:
            self.metrics_summary = metrics_summary
        if name is not None:
            self.name = name
        if number is not None:
            self.number = number
        self.progress_percent = progress_percent
        if show_plots is not None:
            self.show_plots = show_plots
        if status is not None:
            self.status = status
        if status_last_updated is not None:
            self.status_last_updated = status_last_updated
        if status_reason is not None:
            self.status_reason = status_reason
        if termination_protection is not None:
            self.termination_protection = termination_protection
        if updated_dt is not None:
            self.updated_dt = updated_dt

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmExperiment.  # noqa: E501


        :return: The created_dt of this OrmExperiment.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmExperiment.


        :param created_dt: The created_dt of this OrmExperiment.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def distributed_spec(self):
        """Gets the distributed_spec of this OrmExperiment.  # noqa: E501


        :return: The distributed_spec of this OrmExperiment.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._distributed_spec

    @distributed_spec.setter
    def distributed_spec(self, distributed_spec):
        """Sets the distributed_spec of this OrmExperiment.


        :param distributed_spec: The distributed_spec of this OrmExperiment.  # noqa: E501
        :type distributed_spec: dict(str, object)
        """

        self._distributed_spec = distributed_spec

    @property
    def edges(self):
        """Gets the edges of this OrmExperiment.  # noqa: E501


        :return: The edges of this OrmExperiment.  # noqa: E501
        :rtype: OrmExperimentEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmExperiment.


        :param edges: The edges of this OrmExperiment.  # noqa: E501
        :type edges: OrmExperimentEdges
        """

        self._edges = edges

    @property
    def estimated_end_dt(self):
        """Gets the estimated_end_dt of this OrmExperiment.  # noqa: E501


        :return: The estimated_end_dt of this OrmExperiment.  # noqa: E501
        :rtype: datetime
        """
        return self._estimated_end_dt

    @estimated_end_dt.setter
    def estimated_end_dt(self, estimated_end_dt):
        """Sets the estimated_end_dt of this OrmExperiment.


        :param estimated_end_dt: The estimated_end_dt of this OrmExperiment.  # noqa: E501
        :type estimated_end_dt: datetime
        """

        self._estimated_end_dt = estimated_end_dt

    @property
    def exit_code(self):
        """Gets the exit_code of this OrmExperiment.  # noqa: E501


        :return: The exit_code of this OrmExperiment.  # noqa: E501
        :rtype: int
        """
        return self._exit_code

    @exit_code.setter
    def exit_code(self, exit_code):
        """Sets the exit_code of this OrmExperiment.


        :param exit_code: The exit_code of this OrmExperiment.  # noqa: E501
        :type exit_code: int
        """

        self._exit_code = exit_code

    @property
    def experiment_created_by(self):
        """Gets the experiment_created_by of this OrmExperiment.  # noqa: E501


        :return: The experiment_created_by of this OrmExperiment.  # noqa: E501
        :rtype: int
        """
        return self._experiment_created_by

    @experiment_created_by.setter
    def experiment_created_by(self, experiment_created_by):
        """Sets the experiment_created_by of this OrmExperiment.


        :param experiment_created_by: The experiment_created_by of this OrmExperiment.  # noqa: E501
        :type experiment_created_by: int
        """

        self._experiment_created_by = experiment_created_by

    @property
    def experiment_plot_volume(self):
        """Gets the experiment_plot_volume of this OrmExperiment.  # noqa: E501


        :return: The experiment_plot_volume of this OrmExperiment.  # noqa: E501
        :rtype: int
        """
        return self._experiment_plot_volume

    @experiment_plot_volume.setter
    def experiment_plot_volume(self, experiment_plot_volume):
        """Sets the experiment_plot_volume of this OrmExperiment.


        :param experiment_plot_volume: The experiment_plot_volume of this OrmExperiment.  # noqa: E501
        :type experiment_plot_volume: int
        """

        self._experiment_plot_volume = experiment_plot_volume

    @property
    def experiment_project(self):
        """Gets the experiment_project of this OrmExperiment.  # noqa: E501


        :return: The experiment_project of this OrmExperiment.  # noqa: E501
        :rtype: int
        """
        return self._experiment_project

    @experiment_project.setter
    def experiment_project(self, experiment_project):
        """Sets the experiment_project of this OrmExperiment.


        :param experiment_project: The experiment_project of this OrmExperiment.  # noqa: E501
        :type experiment_project: int
        """

        self._experiment_project = experiment_project

    @property
    def experiment_sweep(self):
        """Gets the experiment_sweep of this OrmExperiment.  # noqa: E501


        :return: The experiment_sweep of this OrmExperiment.  # noqa: E501
        :rtype: int
        """
        return self._experiment_sweep

    @experiment_sweep.setter
    def experiment_sweep(self, experiment_sweep):
        """Sets the experiment_sweep of this OrmExperiment.


        :param experiment_sweep: The experiment_sweep of this OrmExperiment.  # noqa: E501
        :type experiment_sweep: int
        """

        self._experiment_sweep = experiment_sweep

    @property
    def id(self):
        """Gets the id of this OrmExperiment.  # noqa: E501


        :return: The id of this OrmExperiment.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmExperiment.


        :param id: The id of this OrmExperiment.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def immutable_slug(self):
        """Gets the immutable_slug of this OrmExperiment.  # noqa: E501


        :return: The immutable_slug of this OrmExperiment.  # noqa: E501
        :rtype: str
        """
        return self._immutable_slug

    @immutable_slug.setter
    def immutable_slug(self, immutable_slug):
        """Sets the immutable_slug of this OrmExperiment.


        :param immutable_slug: The immutable_slug of this OrmExperiment.  # noqa: E501
        :type immutable_slug: str
        """

        self._immutable_slug = immutable_slug

    @property
    def is_distributed(self):
        """Gets the is_distributed of this OrmExperiment.  # noqa: E501


        :return: The is_distributed of this OrmExperiment.  # noqa: E501
        :rtype: bool
        """
        return self._is_distributed

    @is_distributed.setter
    def is_distributed(self, is_distributed):
        """Sets the is_distributed of this OrmExperiment.


        :param is_distributed: The is_distributed of this OrmExperiment.  # noqa: E501
        :type is_distributed: bool
        """

        self._is_distributed = is_distributed

    @property
    def is_local(self):
        """Gets the is_local of this OrmExperiment.  # noqa: E501


        :return: The is_local of this OrmExperiment.  # noqa: E501
        :rtype: bool
        """
        return self._is_local

    @is_local.setter
    def is_local(self, is_local):
        """Sets the is_local of this OrmExperiment.


        :param is_local: The is_local of this OrmExperiment.  # noqa: E501
        :type is_local: bool
        """

        self._is_local = is_local

    @property
    def last_alive(self):
        """Gets the last_alive of this OrmExperiment.  # noqa: E501


        :return: The last_alive of this OrmExperiment.  # noqa: E501
        :rtype: datetime
        """
        return self._last_alive

    @last_alive.setter
    def last_alive(self, last_alive):
        """Sets the last_alive of this OrmExperiment.


        :param last_alive: The last_alive of this OrmExperiment.  # noqa: E501
        :type last_alive: datetime
        """

        self._last_alive = last_alive

    @property
    def local_execution_spec(self):
        """Gets the local_execution_spec of this OrmExperiment.  # noqa: E501


        :return: The local_execution_spec of this OrmExperiment.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._local_execution_spec

    @local_execution_spec.setter
    def local_execution_spec(self, local_execution_spec):
        """Sets the local_execution_spec of this OrmExperiment.


        :param local_execution_spec: The local_execution_spec of this OrmExperiment.  # noqa: E501
        :type local_execution_spec: dict(str, object)
        """

        self._local_execution_spec = local_execution_spec

    @property
    def message(self):
        """Gets the message of this OrmExperiment.  # noqa: E501


        :return: The message of this OrmExperiment.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this OrmExperiment.


        :param message: The message of this OrmExperiment.  # noqa: E501
        :type message: str
        """

        self._message = message

    @property
    def metrics_summary(self):
        """Gets the metrics_summary of this OrmExperiment.  # noqa: E501


        :return: The metrics_summary of this OrmExperiment.  # noqa: E501
        :rtype: dict(str, object)
        """
        return self._metrics_summary

    @metrics_summary.setter
    def metrics_summary(self, metrics_summary):
        """Sets the metrics_summary of this OrmExperiment.


        :param metrics_summary: The metrics_summary of this OrmExperiment.  # noqa: E501
        :type metrics_summary: dict(str, object)
        """

        self._metrics_summary = metrics_summary

    @property
    def name(self):
        """Gets the name of this OrmExperiment.  # noqa: E501


        :return: The name of this OrmExperiment.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this OrmExperiment.


        :param name: The name of this OrmExperiment.  # noqa: E501
        :type name: str
        """

        self._name = name

    @property
    def number(self):
        """Gets the number of this OrmExperiment.  # noqa: E501


        :return: The number of this OrmExperiment.  # noqa: E501
        :rtype: int
        """
        return self._number

    @number.setter
    def number(self, number):
        """Sets the number of this OrmExperiment.


        :param number: The number of this OrmExperiment.  # noqa: E501
        :type number: int
        """

        self._number = number

    @property
    def progress_percent(self):
        """Gets the progress_percent of this OrmExperiment.  # noqa: E501


        :return: The progress_percent of this OrmExperiment.  # noqa: E501
        :rtype: float
        """
        return self._progress_percent

    @progress_percent.setter
    def progress_percent(self, progress_percent):
        """Sets the progress_percent of this OrmExperiment.


        :param progress_percent: The progress_percent of this OrmExperiment.  # noqa: E501
        :type progress_percent: float
        """

        self._progress_percent = progress_percent

    @property
    def show_plots(self):
        """Gets the show_plots of this OrmExperiment.  # noqa: E501


        :return: The show_plots of this OrmExperiment.  # noqa: E501
        :rtype: bool
        """
        return self._show_plots

    @show_plots.setter
    def show_plots(self, show_plots):
        """Sets the show_plots of this OrmExperiment.


        :param show_plots: The show_plots of this OrmExperiment.  # noqa: E501
        :type show_plots: bool
        """

        self._show_plots = show_plots

    @property
    def status(self):
        """Gets the status of this OrmExperiment.  # noqa: E501


        :return: The status of this OrmExperiment.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this OrmExperiment.


        :param status: The status of this OrmExperiment.  # noqa: E501
        :type status: str
        """

        self._status = status

    @property
    def status_last_updated(self):
        """Gets the status_last_updated of this OrmExperiment.  # noqa: E501


        :return: The status_last_updated of this OrmExperiment.  # noqa: E501
        :rtype: datetime
        """
        return self._status_last_updated

    @status_last_updated.setter
    def status_last_updated(self, status_last_updated):
        """Sets the status_last_updated of this OrmExperiment.


        :param status_last_updated: The status_last_updated of this OrmExperiment.  # noqa: E501
        :type status_last_updated: datetime
        """

        self._status_last_updated = status_last_updated

    @property
    def status_reason(self):
        """Gets the status_reason of this OrmExperiment.  # noqa: E501


        :return: The status_reason of this OrmExperiment.  # noqa: E501
        :rtype: str
        """
        return self._status_reason

    @status_reason.setter
    def status_reason(self, status_reason):
        """Sets the status_reason of this OrmExperiment.


        :param status_reason: The status_reason of this OrmExperiment.  # noqa: E501
        :type status_reason: str
        """

        self._status_reason = status_reason

    @property
    def termination_protection(self):
        """Gets the termination_protection of this OrmExperiment.  # noqa: E501


        :return: The termination_protection of this OrmExperiment.  # noqa: E501
        :rtype: bool
        """
        return self._termination_protection

    @termination_protection.setter
    def termination_protection(self, termination_protection):
        """Sets the termination_protection of this OrmExperiment.


        :param termination_protection: The termination_protection of this OrmExperiment.  # noqa: E501
        :type termination_protection: bool
        """

        self._termination_protection = termination_protection

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmExperiment.  # noqa: E501


        :return: The updated_dt of this OrmExperiment.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmExperiment.


        :param updated_dt: The updated_dt of this OrmExperiment.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmExperiment):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmExperiment):
            return True

        return self.to_dict() != other.to_dict()



import os
import zipfile

from vessl.util import logger


class Zipper(zipfile.ZipFile):
    def __init__(self, file, mode):
        super(Zipper, self).__init__(file, mode)

    @classmethod
    def zip(cls, path, ziph):
        if not os.path.isabs(path):
            path = os.path.isabs(path)
        if os.path.isdir(path):
            cls.zipdir(path, ziph)
        elif os.path.isfile(path):
            cls.zipfile(path, ziph)

    @classmethod
    def zipfile(cls, path, ziph):
        ziph.write(path, os.path.basename(path))

    @classmethod
    def zipdir(cls, path, ziph):
        for root, dirs, files in os.walk(path):
            files = [f for f in files if not f.startswith(".")]
            dirs[:] = [d for d in dirs if not d.startswith(".")]
            for file in files:
                filename = os.path.join(root, file)
                ziph.write(filename, os.path.relpath(os.path.join(root, file), path))
                logger.debug(f"Compressed {filename}.")



import os
import tempfile
from typing import List, Optional

from openapi_client.models import (
    ProtoVolumeMountRequest,
    ProtoVolumeMountRequests,
    ProtoVolumeMountRequestSourceArchiveFile,
    ProtoVolumeMountRequestSourceCode,
    ProtoVolumeMountRequestSourceDataset,
    ProtoVolumeMountRequestSourceDatasetVersion,
    ResponseFileMetadata,
    VolumeFileCopyAPIPayload,
    VolumeFileCreateAPIPayload,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.project import _get_project
from vessl.util.constant import (
    DATASET_VERSION_HASH_LATEST,
    MOUNT_PATH_EMPTY_DIR,
    MOUNT_TYPE_ARCHIVE_FILE,
    MOUNT_TYPE_CODE,
    MOUNT_TYPE_DATASET,
    MOUNT_TYPE_DATASET_VERSION,
    MOUNT_TYPE_EMPTY_DIR,
    MOUNT_TYPE_OUTPUT,
)
from vessl.util.downloader import Downloader
from vessl.util.exception import GitError, InvalidVolumeFileError, VesslApiException
from vessl.util.git import get_git_diff_path, get_git_ref, get_git_repo
from vessl.util.random import random_string
from vessl.util.uploader import Uploader
from vessl.util.zipper import Zipper


def read_volume_file(volume_id: int, path: str) -> ResponseFileMetadata:
    return vessl_api.volume_file_read_api(volume_id=volume_id, path=path)


def list_volume_files(
    volume_id: int,
    need_download_url: bool = False,
    path: str = "",
    recursive: bool = False,
) -> List[ResponseFileMetadata]:
    result = vessl_api.volume_file_list_api(
        volume_id=volume_id,
        recursive=recursive,
        path=path,
        need_download_url=need_download_url,
    ).results
    return sorted(result, key=lambda x: x.path)


def create_volume_file(volume_id: int, is_dir: bool, path: str) -> ResponseFileMetadata:
    result = vessl_api.volume_file_create_api(
        volume_id=volume_id,
        volume_file_create_api_payload=VolumeFileCreateAPIPayload(
            is_dir=is_dir,
            path=path,
        ),
    )
    assert isinstance(result, ResponseFileMetadata)
    return result


def delete_volume_file(
    volume_id: int, path: str, recursive: bool = False
) -> List[ResponseFileMetadata]:
    return vessl_api.volume_file_delete_api(
        volume_id=volume_id, path=path, recursive=recursive
    ).deleted_files


def upload_volume_file(volume_id: int, path: str) -> ResponseFileMetadata:
    return vessl_api.volume_file_uploaded_api(volume_id=volume_id, path=path)


def copy_volume_file(
    source_volume_id: Optional[int],
    source_path: str,
    dest_volume_id: Optional[int],
    dest_path: str,
    recursive: bool = False,
) -> Optional[List[ResponseFileMetadata]]:
    if source_volume_id is None and dest_volume_id:
        return _copy_volume_file_local_to_remote(
            source_path,
            dest_volume_id,
            dest_path,
        )

    if dest_volume_id is None and source_volume_id:
        return _copy_volume_file_remote_to_local(
            source_volume_id,
            source_path,
            dest_path,
        )

    if source_volume_id and dest_volume_id:
        return _copy_volume_file_remote_to_remote(
            source_volume_id, source_path, dest_volume_id, dest_path, recursive
        )


def _copy_volume_file_local_to_remote(
    source_path: str, dest_volume_id: int, dest_path: str
) -> List[ResponseFileMetadata]:
    """Copy local to remote

    Behavior works like linux cp command
    - `source_path` is file
      - `dest_path` is not a directory: copy as file with new name
      - `dest_path` is directory: copy file into directory with original name
    - `source_path` is directory
      - `dest_path` is file: error
      - `dest_path` does not exist: create `dest_path` and copy contents of `source_path`
      - `dest_path` exists: copy `source_path` as subdirectory of `dest_path`
    """

    output = "Successfully uploaded {} out of {} file(s)."
    source_path = source_path.rstrip("/")

    try:
        dest_file = read_volume_file(dest_volume_id, dest_path)
    except VesslApiException:
        dest_file = None

    if not os.path.isdir(source_path):
        if dest_file and (dest_file.is_dir or dest_path.endswith("/")):
            dest_path = os.path.join(dest_path, os.path.basename(source_path))

        uploaded_file = Uploader.upload(source_path, dest_volume_id, dest_path)

        print(output.format(1, 1))
        return [uploaded_file]

    if dest_file and not dest_file.is_dir:
        raise InvalidVolumeFileError(
            f"Destination path is not a directory: {dest_path}."
        )

    if dest_file and dest_file.is_dir:
        dest_path = os.path.join(dest_path, os.path.basename(source_path))

    paths = Uploader.get_paths_in_dir(source_path)
    uploaded_files = Uploader.bulk_upload(source_path, paths, dest_volume_id, dest_path)

    print(output.format(len(uploaded_files), len(paths)))
    return uploaded_files


def _copy_volume_file_remote_to_local(
    source_volume_id: int,
    source_path: str,
    dest_path: str,
) -> None:
    """Copy remote to local

    Behavior works like linux cp command
    - `source_path` is file
      - `dest_path` is not a directory: copy as file with new name
      - `dest_path` is directory: copy file into directory with original name
    - `source_path` is directory
      - `dest_path` is file: error
      - `dest_path` does not exist: create `dest_path` and copy contents of `source_path`
      - `dest_path` exists: copy `source_path` as subdirectory of `dest_path`
    """

    output = "Successfully downloaded {} out of {} file(s)."
    source_file = read_volume_file(source_volume_id, source_path)

    if not source_file.is_dir:
        if os.path.isdir(dest_path):
            dest_path = os.path.join(dest_path, os.path.basename(source_file.path))

        Downloader.download(dest_path, source_file)

        print(output.format(1, 1))
        return

    files = list_volume_files(
        volume_id=source_volume_id,
        need_download_url=True,
        path=source_path,
        recursive=True,
    )

    if os.path.isfile(dest_path):
        raise InvalidVolumeFileError(
            f"Destination path is not a directory: {dest_path}."
        )

    prefix = source_path.rstrip("/")
    if os.path.isdir(dest_path):
        prefix = os.path.dirname(prefix)

    if prefix:
        prefix += "/"
        for file in files:
            file.path = file.path.replace(prefix, "", 1)  # Remove prefix
            file.path = os.path.normpath(file.path)

    downloaded_files = Downloader.bulk_download(dest_path, files)
    print(output.format(len(downloaded_files), len([x for x in files if not x.is_dir])))


def _copy_volume_file_remote_to_remote(
    source_volume_id: int,
    source_path: str,
    dest_volume_id: int,
    dest_path: str,
    recursive: bool,
) -> None:
    if source_volume_id != dest_volume_id:
        raise InvalidVolumeFileError("Files can only be copied within the same volume.")

    files = vessl_api.volume_file_copy_api(
        volume_id=source_volume_id,
        volume_file_copy_api_payload=VolumeFileCopyAPIPayload(
            dest_path=dest_path,
            recursive=recursive,
            source_dataset_version="latest",
            source_path=source_path,
        ),
    ).copied_files

    print(f"Successfully downloaded {len(files)} out of {len(files)} file(s).")


def _configure_volume_mount_requests(
    dataset_mounts: Optional[List[str]],
    git_ref_mounts: Optional[List[str]],
    git_diff_mount: Optional[str],
    local_files: Optional[List[str]],
    upload_local_git_diff: bool,
    archive_file_mount: Optional[str],
    root_volume_size: Optional[str],
    working_dir: Optional[str],
    output_dir: Optional[str],
    **kwargs,
) -> ProtoVolumeMountRequests:
    requests = [
        _configure_volume_mount_request_empty_dir(),
        _configure_volume_mount_request_output(output_dir),
    ]

    if local_files:
        requests.extend(_configure_volume_mount_request_local_files(local_files, **kwargs))

    if not git_ref_mounts and git_diff_mount is None and archive_file_mount is None:
        # No information for code mount is given - user is creating new experiment from CLI.
        # Generate VolumeMountRequest from projectRepository and local working directory
        requests.extend(_configure_volume_mount_request_local_git(upload_local_git_diff, **kwargs))
    else:
        # Explicit information for code mount is given - user is reproducing experiment from CLI.
        if not git_ref_mounts:
            requests.extend(
                _configure_volume_mount_request_codes(
                    git_ref_mounts, git_diff_mount, **kwargs
                )
            )
        if archive_file_mount is not None:
            requests.append(
                _configure_volume_mount_request_archive_file(
                    archive_file_mount, **kwargs
                )
            )

    if dataset_mounts is not None:
        requests += [
            _configure_volume_mount_request_dataset(dataset_mount, **kwargs)
            for dataset_mount in dataset_mounts
        ]

    return ProtoVolumeMountRequests(
        root_volume_size=root_volume_size,
        working_dir=working_dir,
        requests=requests,
    )


def _configure_volume_mount_request_dataset(
    dataset_mount: str, **kwargs
) -> ProtoVolumeMountRequest:
    from vessl.dataset import read_dataset, read_dataset_version

    mount_path, dataset_path = dataset_mount.split(":", 1)
    mount_path = os.path.join(mount_path, "")  # Ensure path ends in /

    organization_name = _get_organization_name(**kwargs)
    dataset_name = dataset_path
    dataset_version_hash = DATASET_VERSION_HASH_LATEST

    if "@" in dataset_path:
        # Example: mnist@3d1e0f91c
        dataset_name, dataset_version_hash = dataset_path.rsplit("@", 1)

    if "/" in dataset_name:
        # Example: org1/mnist@3d1e0f91c
        organization_name, dataset_name = dataset_name.split("/", 1)

    dataset = read_dataset(dataset_name, organization_name=organization_name)

    if not dataset.is_version_enabled:
        return ProtoVolumeMountRequest(
            mount_type=MOUNT_TYPE_DATASET,
            mount_path=mount_path,
            dataset=ProtoVolumeMountRequestSourceDataset(
                dataset_id=dataset.id,
                dataset_name=dataset_name,
            ),
        )

    if dataset_version_hash != DATASET_VERSION_HASH_LATEST:
        dataset_version_hash = read_dataset_version(
            dataset.id, dataset_version_hash
        ).version_hash  # Get full version hash from truncated one given by input

    return ProtoVolumeMountRequest(
        mount_type=MOUNT_TYPE_DATASET_VERSION,
        mount_path=mount_path,
        dataset_version=ProtoVolumeMountRequestSourceDatasetVersion(
            dataset_id=dataset.id,
            dataset_name=dataset_name,
            dataset_version_hash=dataset_version_hash,
        ),
    )


def _configure_volume_mount_request_local_files(local_files: List[str], **kwargs) -> List[ProtoVolumeMountRequest]:
    project = _get_project(**kwargs)

    volume_mount_requests: List[ProtoVolumeMountRequest] = []
    for local_file in local_files:
        if ":" in local_file:
            local_path, remote_path = local_file.split(":")
            if not os.path.isabs(remote_path):
                remote_path = os.path.join("/home/vessl", remote_path)
        else:
            local_path = local_file
            remote_path = "/home/vessl/local"

        cwd = os.path.abspath(local_path)
        zip_path = os.path.abspath(
            os.path.join(tempfile.gettempdir(), f"{project.name}_{random_string()}.zip")
        )
        zipper = Zipper(zip_path, "w")
        zipper.zip(cwd, zipper)
        zipper.close()

        uploaded = Uploader.upload(
            local_path=zipper.filename,
            volume_id=project.volume_id,
            remote_path=os.path.basename(zipper.filename),
        )
        os.remove(zipper.filename)
        volume_mount_requests.append(
            _configure_volume_mount_request_archive_file(
                archive_file_mount=f"{remote_path}:{uploaded.path}"
            )
        )

    return volume_mount_requests


def _configure_volume_mount_request_local_git(upload_local_git_diff, **kwargs) -> List[ProtoVolumeMountRequest]:
    project = _get_project(**kwargs)

    local_git_owner, local_git_repo = None, None
    try:
        local_git_owner, local_git_repo = get_git_repo()
    except GitError:
        pass

    volume_mount_requests: List[ProtoVolumeMountRequest] = []
    used_mount_paths = set()
    for repo in project.project_repositories:
        mount_path = f"/home/vessl/{repo.git_repo}"
        if mount_path in used_mount_paths:
            mount_path = f"/home/vessl/{repo.git_repo}-{repo.git_owner}"

        git_mount_info = f"github/{repo.git_owner}/{repo.git_repo}/latest"
        git_diff_path = None
        if upload_local_git_diff and repo.git_owner == local_git_owner and repo.git_repo == local_git_repo:
            # Current working directory is one of the project repository.
            # Mount VolumeSourceCode with potential uncommitted changes
            git_diff_path = get_git_diff_path(project)
            git_mount_info = git_mount_info.replace("latest", get_git_ref())
        volume_mount_requests.append(
            _configure_volume_mount_request_code(
                git_ref_mount=f"{mount_path}:{git_mount_info}",
                git_diff_mount=None
                if git_diff_path is None
                else f"{mount_path}:{git_diff_path}",
            )
        )

    return volume_mount_requests


def _configure_volume_mount_request_codes(
    git_ref_mounts: List[str], git_diff_mount: str, **kwargs
) -> List[ProtoVolumeMountRequest]:
    return [
        _configure_volume_mount_request_code(git_ref_mount, git_diff_mount, **kwargs)
        for git_ref_mount in git_ref_mounts
    ]


def _configure_volume_mount_request_code(
    git_ref_mount: str, git_diff_mount: str, **kwargs
) -> ProtoVolumeMountRequest:
    # Generate ProtoVolumeMountRequestSourceCode from git_ref_mount
    # ref_mount_info = <provider>/<owner>/<repo>/<commit>
    ref_mount_path, ref_mount_info = git_ref_mount.split(":", 1)
    code_info = ref_mount_info.split("/", 3)

    vmr_source_code = ProtoVolumeMountRequestSourceCode(
        git_provider=code_info[0],
        git_owner=code_info[1],
        git_repo=code_info[2],
        git_ref=code_info[3],
    )

    # Add optional git_diff parameter if git_diff_mount has same mount path
    if git_diff_mount is not None:
        diff_mount_path, diff_mount_file = git_diff_mount.split(":", 1)
        if ref_mount_path == diff_mount_path:
            vmr_source_code.git_diff_file = diff_mount_file

    return ProtoVolumeMountRequest(
        mount_type=MOUNT_TYPE_CODE,
        mount_path=ref_mount_path,
        code=vmr_source_code,
    )


def _configure_volume_mount_request_archive_file(
    archive_file_mount: str,
) -> ProtoVolumeMountRequest:
    mount_path, archive_file_path = archive_file_mount.split(":", 1)

    return ProtoVolumeMountRequest(
        mount_type=MOUNT_TYPE_ARCHIVE_FILE,
        mount_path=mount_path,
        archive_file=ProtoVolumeMountRequestSourceArchiveFile(
            archive_file=archive_file_path,
        ),
    )


def _configure_volume_mount_request_empty_dir() -> ProtoVolumeMountRequest:
    return ProtoVolumeMountRequest(
        mount_type=MOUNT_TYPE_EMPTY_DIR,
        mount_path=MOUNT_PATH_EMPTY_DIR,
    )


def _configure_volume_mount_request_output(output_dir: str) -> ProtoVolumeMountRequest:
    return ProtoVolumeMountRequest(
        mount_type=MOUNT_TYPE_OUTPUT,
        mount_path=output_dir,
    )



import json

from openapi_client.exceptions import ApiException
from vessl._version import __VERSION__

DEFAULT_ERROR_MESSAGE = (
    "An unexpected exception occurred. Use VESSL_LOG=DEBUG to view stack trace. "
    "(CLI version: %s)\n"
    "(`pip install --upgrade vessl` might resolve this issue.)" % __VERSION__
)


class VesslException(Exception):
    def __init__(self, message=DEFAULT_ERROR_MESSAGE, exit_code=1):
        self.message = message
        self.exit_code = exit_code
        super().__init__(message)


class VesslApiException(VesslException):
    @classmethod
    def convert_api_exception(cls, api_exception: ApiException) -> "VesslApiException":
        try:
            body = json.loads(api_exception.body)
        except json.JSONDecodeError:
            return cls()

        body_code = body.get("code")
        body_message = body.get("message", "")
        message = (
            f"{body_code} ({api_exception.status})"
            f"{': ' + body_message if body_message else ''}"
        )

        fields = body.get("fields")
        if fields:
            additional_messages = []
            for field in fields:
                field_name = field.get("name", "")
                field_value = field.get("value", "")
                field_message = field.get("message")
                additional_messages.append(
                    f"{field_name}: {field_value}"
                    f"{'(' + field_message + ')' if field_message else ''}"
                )
            message += f" {', '.join(additional_messages)}."

        return cls(message=message)


class GitError(VesslException):
    pass


class TimeoutError(VesslException):
    pass


class InvalidDatasetError(VesslException):
    pass


class InvalidKernelClusterError(VesslException):
    pass


class InvalidKernelImageError(VesslException):
    pass


class InvalidKernelResourceSpecError(VesslException):
    pass


class InvalidOrganizationError(VesslException):
    pass


class InvalidExperimentError(VesslException):
    pass


class InvalidProjectError(VesslException):
    pass


class InvalidTokenError(VesslException):
    pass


class InvalidVolumeFileError(VesslException):
    pass


class InvalidVolumeMountError(VesslException):
    pass


class InvalidWorkspaceError(VesslException):
    pass


class InvalidParamsError(VesslException):
    pass


class InvalidTypeError(VesslException):
    pass


class ImportPackageError(VesslException):
    pass



from test.conftest import USE_MOCK
from test.util import random_string

import pytest
from mock import Mock, patch

import vessl
from vessl.util.exception import InvalidOrganizationError


@pytest.mark.skipif(USE_MOCK, reason="Does not run if mocking is used.")
class TestOrganization:
    organization_name = random_string()

    @pytest.mark.order(index=1)
    def test_create_organization(self):
        vessl.create_organization(
            organization_name=self.organization_name,
            region=vessl.vessl_api.region_list_api().default_region,
        )

    def test_read_organization(self):
        vessl.read_organization(self.organization_name)


def test_list_organizations():
    vessl.list_organizations()


def test_get_organization_name():
    with patch.object(vessl.vessl_api, "organization", None):
        with pytest.raises(InvalidOrganizationError):
            vessl.organization._get_organization_name()

    organization_name = "org"
    assert (
        vessl.organization._get_organization_name(organization_name=organization_name)
        == organization_name
    )

    organization = Mock()
    organization.name = organization_name
    with patch.object(vessl.vessl_api, "organization", organization):
        assert vessl.organization._get_organization_name() == organization_name
